from pydantic import Field, BaseModel

from retk.const import settings, USER_BEHAVIOR_TYPE_MAP

_max_len_behavior_type = len(USER_BEHAVIOR_TYPE_MAP)


class UserBehaviorRequest(BaseModel):
    type: int = Field(le=_max_len_behavior_type)
    remark: str = Field(default="", max_length=settings.MAX_STATISTIC_REMARK_LENGTH)



import functools
import os
import re
import unicodedata

import bson
import sortedcontainers

from .errors import MongitaError

ASCENDING = 1
DESCENDING = -1

_windows_device_files = ('CON', 'AUX', 'COM1', 'COM2', 'COM3', 'COM4', 'LPT1',
                         'LPT2', 'LPT3', 'PRN', 'NUL')
_filename_ascii_strip_re = re.compile(r'[^A-Za-z0-9_.-]')
_invalid_names = re.compile(r'[/\. "$*<>:|?]')


def secure_filename(filename: str) -> str:
    """
    The idea of this is to ensure that the document_id doesn't do sketchy shit on
    the filesystem. This will probably be deleted soon.
    """
    r"""From werkzeug source"""
    filename = unicodedata.normalize("NFKD", filename)
    filename = filename.encode("ascii", "ignore").decode("ascii")
    for sep in os.path.sep, os.path.altsep:
        if sep:
            filename = filename.replace(sep, " ")
    filename = str(_filename_ascii_strip_re.sub("", "_".join(filename.split()))).strip(
        "._"
    )
    # on nt a couple of special files are present in each folder.  We
    # have to ensure that the target file is not such a filename.  In
    # this case we prepend an underline
    if filename and filename.split(".")[0].upper() in _windows_device_files:
        filename = f"_{filename}"
    return filename


def ok_name(name):
    """
    In-line with MongoDB restrictions.
    https://docs.mongodb.com/manual/reference/limits/#std-label-restrictions-on-db-names
    https://docs.mongodb.com/manual/reference/limits/#Restriction-on-Collection-Names
    The prohibition on "system." names will be covered by the prohibition on '.'
    """
    if not name:
        return False
    if _invalid_names.search(name):
        return False
    if len(name) > 64:
        return False
    return True


def support_alert(func):
    """
    Provide smart tips if the user tries to use un-implemented / deprecated
    known kwargs.
    """

    @functools.wraps(func)
    def inner(*args, **kwargs):
        for k in kwargs:
            if k not in func.__code__.co_varnames:
                raise MongitaError("The argument %r is not supported by %r in Mongita. "
                                   "This may or may not be supported in PyMongo. "
                                   "If it is, you can help implement it." %
                                   (k, func))
        return func(*args, **kwargs)

    return inner


class MetaStorageObject(dict):
    """
    Subclass of the StorageObject with some extra handling for metadata.
    Specifically, indexes need extra steps to fully encode / decode.
    """

    def __init__(self, doc):
        super().__init__(doc)

    def to_storage(self, as_bson=False):
        """
        Makes sure that the SortedDict indexes are bson-compatible
        """
        if as_bson:
            if 'indexes' in self:
                swap = {}
                self_indexes = self['indexes']
                for idx_key in self_indexes.keys():
                    swap[idx_key] = self_indexes[idx_key]['idx']
                    idx = map(lambda tup: (tup[0], list(tup[1])),
                              self_indexes[idx_key]['idx'].items())
                    self_indexes[idx_key]['idx'] = list(idx)
                ret = bson.encode(self)
                for idx_key, idx in swap.items():
                    self_indexes[idx_key]['idx'] = idx
                return ret
            return bson.encode(self)
        return self

    @staticmethod
    def from_storage(obj, from_bson=False):
        if from_bson:
            doc = bson.decode(obj)
            so = MetaStorageObject(doc)
            so.decode_indexes()
            return so
        return obj

    def decode_indexes(self):
        """
        Changes the encoded indexes to SortedDicts
        """
        if 'indexes' in self:
            self_indexes = self['indexes']
            for idx_key in self_indexes.keys():
                idx = list(map(lambda tup: (tuple(tup[0]), set(tup[1])),
                               self['indexes'][idx_key]['idx']))
                self_indexes[idx_key]['idx'] = sortedcontainers.SortedDict(idx)



"""
This is the base class for all plugins. It is used to define the interface for all plugins.
The plugin can only be used in self-hosted mode,
because every customized plugin consumes resources.


Usage:

1. Create a new plugin by inheriting from `Plugin`:

    ```python
    import retk

    class MyPlugin(retk.Plugin):
        name = "MyPlugin"
        version = "0.1.0"
        description = "A demo plugin."
        author = "me"


        def before_node_updated(self, uid: str, nid: str, data: Dict[str, Any]) -> None:
            print("before_node_updated")
            data["md"] = "before_node_updated:" + data["md"]
    ```

2. Add the plugin to the system:

    ```python
    retk.add_plugin(MyPlugin())
    ```

3. Run rethink server:

    ```python
    retk.run()
    ```

4. Check when update a node, the `before_node_updated` text will show on your note.
"""

import base64
import inspect
import os
from dataclasses import dataclass
from typing import Dict, List, Optional, Any

from retk import const
from retk.core.scheduler.timing import Timing
from retk.models import tps


@dataclass
class PluginAPICallReturn:
    success: bool
    message: str
    data: Any


class Plugin:
    """
    Attributes:
        name (str): The name of the plugin.
        version (str): The version of the plugin.
        description (str): The description of the plugin.
        author (str): The author of the plugin.
        icon (str): The icon path for this plugin.
        activated (bool): Whether the plugin is activated.
        schedule_timing (Optional[Timing]): The schedule timing for the plugin if `on_schedule` method has been defined.
    """
    id: str
    name: str
    version: str = ""
    description: str = ""
    author: str = ""
    icon: str = ""
    activated: bool = True
    schedule_timing: Optional[Timing] = None

    def __init__(self):
        for attr in ["id", "name", "version", "description", "author", "icon"]:
            if not hasattr(self, attr):
                raise ValueError(f"attribute '{attr}' is required")
        if len(self.id) == 0:
            raise ValueError("id is required")
        if len(self.id) > const.settings.PLUGIN_ID_MAX_LENGTH:
            raise ValueError(f"id length should be less than {const.settings.PLUGIN_ID_MAX_LENGTH}")

        # get icon and convert to base64
        if os.path.isfile(self.icon):
            icon_path = self.icon
        else:
            pd = os.path.dirname(inspect.getfile(self.__class__))
            icon_path = os.path.join(pd, str(self.icon))
        with open(icon_path, "rb") as f:
            b64 = base64.b64encode(f.read()).decode()
            ext = str(self.icon).rsplit(".", 1)[-1]
            if ext == "svg":
                ext = "svg+xml"
            elif ext == "jpg":
                ext = "jpeg"
            self.icon_src = f"data:image/{ext};base64,{b64}"

    def render_plugin_home(self, language: str) -> str:
        raise NotImplementedError

    def render_editor_side(self, uid: str, nid: str, md: str, language: str) -> str:
        raise NotImplementedError

    def on_node_added(self, node: tps.Node) -> None:
        raise NotImplementedError

    def before_node_updated(self, uid: str, nid: str, data: Dict[str, Any]) -> None:
        raise NotImplementedError

    def on_node_updated(self, node: tps.Node, old_node: tps.Node) -> None:
        raise NotImplementedError

    def on_schedule(self) -> None:
        raise NotImplementedError

    def handle_api_call(self, method: str, data: Any) -> PluginAPICallReturn:
        """
        api call from your plugin rendered page. etc. plugin home page, editor side bar.

        POST: https://ip:port/plugin/call

        The exact POST url can be found in self.api_call_url

        Body: {
            "pluginId: "your plugin id",
            "method": "your method",
            "data": any data format,
            "requestId": "unique request id string"
        }
        Response: {
            "success": boolean,
            "message": "message",
            "requestId": "unique request id string",
            "pluginId: "your plugin id",
            "method": "your method",
            "data": any data format,
        }

        When your rendered page calls you this api, you will receive

        Args:
            method (str): your customized method name
            data (Any): any data format you need to process

        Returns:
            Any: the result data shows in your Response["data"]
        """
        raise NotImplementedError

    @property
    def api_call_url(self) -> str:
        """
        in your template html, you can pass this url to your javascript code to connect self.handle_api_call() method.

        Returns:
            str: the api call url
        """
        try:
            addr = os.environ["VUE_APP_API_URL"]
        except KeyError:
            raise ValueError("the host or port number is not set in the environment.")
        return f"{addr}/api/plugins/call"


event_plugin_map: Dict[str, List[Plugin]] = {
    "on_node_added": [],
    "before_node_updated": [],
    "on_node_updated": [],
    "on_schedule": [],
    "on_toolbar_click": [],
    "render_plugin_home": [],
    "render_editor_side": [],
}
_plugins: Dict[str, Plugin] = {}


def add_plugin(plugin: Plugin):
    if plugin.id in _plugins:
        raise ValueError(f"plugin '{plugin.id}' already exists")
    _plugins[plugin.id] = plugin

    for method_name, value in event_plugin_map.items():
        cls = plugin.__class__
        method = getattr(cls, method_name, None)
        if method is None:
            continue

        for base_class in inspect.getmro(cls)[1:]:
            if method_name in base_class.__dict__ \
                    and method != base_class.__dict__[method_name]:  # redefined
                if plugin in value:
                    continue
                if method_name == "on_schedule" and plugin.schedule_timing is None:
                    raise ValueError(
                        f"schedule_timing is required for plugin.id={plugin.id} which implements on_schedule method."
                    )
                value.append(plugin)


def get_plugins() -> Dict[str, Plugin]:
    return _plugins


def remove_plugin(plugin: Plugin):
    del _plugins[plugin.id]

    for method_name, value in event_plugin_map.items():
        cls = plugin.__class__
        method = getattr(cls, method_name, None)
        if method is None:
            continue

        for base_class in inspect.getmro(cls)[1:]:
            if method_name in base_class.__dict__ \
                    and method != base_class.__dict__[method_name]:  # redefined
                try:
                    value.remove(plugin)
                except ValueError:
                    pass



from typing import List, Dict, Tuple, Sequence, Literal

from retk.controllers.schemas.node import NodesSearchResponse
from retk.core.recent import put_recent_search
from retk.models import tps
from retk.models.client import client
from retk.models.search_engine.engine import SearchResult
from retk.utils import datetime2str


async def _2node_data(
        hits: Sequence[SearchResult],
) -> List[NodesSearchResponse.Data.Node]:
    nodes = await client.coll.nodes.find({"id": {"$in": [hit.nid for hit in hits]}}).to_list(length=None)
    nodes_map: Dict[str, tps.Node] = {n["id"]: n for n in nodes}
    results = []
    for hit in hits:
        n = nodes_map[hit.nid]
        r = NodesSearchResponse.Data.Node(
            id=n["id"],
            title=n["title"],
            snippet=n["snippet"],
            titleHighlight=hit.titleHighlight,
            bodyHighlights=hit.bodyHighlights,
            score=hit.score,
            type=n["type"],
            createdAt=datetime2str(n["_id"].generation_time),
            modifiedAt=datetime2str(n["modifiedAt"]),
        )
        results.append(r)
    return results


async def user_nodes(
        au: tps.AuthedUser,
        query: str,
        sort_key: Literal[
            "createdAt", "modifiedAt", "title", "similarity"
        ],
        reverse: bool,
        page: int,
        limit: int,
        exclude_nids: Sequence[str],
) -> Tuple[List[NodesSearchResponse.Data.Node], int]:
    # search nodes
    hits, total = await client.search.search(
        au=au,
        query=query,
        sort_key=sort_key,
        reverse=reverse,
        page=page,
        limit=limit,
        exclude_nids=exclude_nids,
    )
    results = await _2node_data(hits)

    if query != "":
        await put_recent_search(au=au, query=query)
    return results, total


async def recommend(
        au: tps.AuthedUser,
        content: str,
        max_return: int = 5,
        exclude_nids: Sequence[str] = None,
) -> List[NodesSearchResponse.Data.Node]:
    if content == "":
        return []
    # search nodes
    hits = await client.search.recommend(
        au=au,
        content=content,
        max_return=max_return,
        exclude_nids=exclude_nids,
    )
    return await _2node_data(hits)


async def at(
        au: tps.AuthedUser,
        nid: str,
        query: str,
        page: int,
        limit: int,
) -> Tuple[List[NodesSearchResponse.Data.Node], int]:
    query = query.strip()

    # if query == "", return recent nodes
    if query == "":
        rn = au.u.last_state.recent_cursor_search_selected_nids
        try:
            rn.remove(nid)
        except ValueError:
            pass
        _nodes = sorted(
            await client.coll.nodes.find({"id": {"$in": rn}}).to_list(length=None), key=lambda x: rn.index(x["id"])
        )
        return [
            NodesSearchResponse.Data.Node(
                id=n["id"],
                title=n["title"],
                snippet=n["snippet"][:60] + "...",
                titleHighlight=n["title"],
                bodyHighlights=[n["snippet"][:60] + "..."],
                score=0,
                type=n["type"],
                createdAt=datetime2str(n["_id"].generation_time),
                modifiedAt=datetime2str(n["modifiedAt"]),
            ) for n in _nodes
        ], len(rn)

    return await user_nodes(
        au=au,
        query=query,
        sort_key="similarity",
        reverse=True,
        page=page,
        limit=limit,
        exclude_nids=[nid],
    )



from retk import core, const
from retk.controllers import schemas
from retk.controllers.node.node_ops import _get_node_search_response_data
from retk.controllers.utils import maybe_raise_json_exception
from retk.models.tps import AuthedUser


async def move_to_trash(
        au: AuthedUser,
        nid: str,
) -> schemas.RequestIdResponse:
    return await move_batch_to_trash(
        au=au,
        req=schemas.node.BatchNodeIdsRequest(
            nids=[nid],
        ),
    )


async def move_batch_to_trash(
        au: AuthedUser,
        req: schemas.node.BatchNodeIdsRequest,
) -> schemas.RequestIdResponse:
    code = await core.node.batch_to_trash(au=au, nids=req.nids)
    maybe_raise_json_exception(au=au, code=code)

    await core.statistic.add_user_behavior(
        uid=au.u.id,
        type_=const.UserBehaviorTypeEnum.NODE_TRASHED_OPS,
        remark="",
    )
    return schemas.RequestIdResponse(
        requestId=au.request_id,
    )


async def get_from_trash(
        au: AuthedUser,
        p: int = 0,
        limit: int = 10,
) -> schemas.node.NodesSearchResponse:
    nodes, total = await core.node.get_nodes_in_trash(au=au, page=p, limit=limit)
    return schemas.node.NodesSearchResponse(
        requestId=au.request_id,
        data=_get_node_search_response_data(nodes=nodes, total=total),
    )


async def restore_from_trash(
        au: AuthedUser,
        nid: str,
) -> schemas.RequestIdResponse:
    return await restore_batch_from_trash(
        au=au,
        req=schemas.node.BatchNodeIdsRequest(
            nids=[nid],
        ),
    )


async def restore_batch_from_trash(
        au: AuthedUser,
        req: schemas.node.BatchNodeIdsRequest,
) -> schemas.RequestIdResponse:
    code = await core.node.restore_batch_from_trash(au=au, nids=req.nids)
    maybe_raise_json_exception(au=au, code=code)

    await core.statistic.add_user_behavior(
        uid=au.u.id,
        type_=const.UserBehaviorTypeEnum.NODE_RESTORED_OPS,
        remark="",
    )
    return schemas.RequestIdResponse(
        requestId=au.request_id,
    )


async def delete_node(
        au: AuthedUser,
        nid: str,
) -> schemas.RequestIdResponse:
    return await delete_batch_node(
        au=au,
        req=schemas.node.BatchNodeIdsRequest(
            nids=[nid],
        ),
    )


async def delete_batch_node(
        au: AuthedUser,
        req: schemas.node.BatchNodeIdsRequest,
) -> schemas.RequestIdResponse:
    code = await core.node.batch_delete(au=au, nids=req.nids)
    maybe_raise_json_exception(au=au, code=code)

    await core.statistic.add_user_behavior(
        uid=au.u.id,
        type_=const.UserBehaviorTypeEnum.NODE_DELETED_OPS,
        remark="",
    )
    return schemas.RequestIdResponse(
        requestId=au.request_id,
    )



import asyncio
import multiprocessing
import traceback
from typing import Optional

from retk.logger import logger
from .obsidian.task import upload_obsidian_task
from .text.task import update_text_task

__process: Optional[multiprocessing.Process] = None
__ctx = multiprocessing.get_context('spawn')
QUEUE = __ctx.Queue()


def __async_task(queue: multiprocessing.Queue):
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.get_event_loop()
    while True:
        item = queue.get()
        task = item.pop("task")
        if task == "md":
            func = update_text_task
        elif task == "obsidian":
            func = upload_obsidian_task
        else:
            logger.error(f"unknown task: {task}")
            continue
        try:
            loop.run_until_complete(func(**item))
        except Exception:  # pylint: disable=broad-except
            msg = traceback.format_exc()
            oneline = msg.replace("\n", "\\n")
            logger.error(f"async task error: {oneline}")
    loop.close()


def init():
    global __process
    __process = __ctx.Process(
        target=__async_task,
        args=(QUEUE,),
        daemon=True,
    )
    __process.start()
    logger.info("async task process stopped")


def stop():
    global __process
    if __process is None:
        return
    __process.terminate()
    __process.join()
    __process = None
    logger.info("async task process stopped")



from datetime import datetime
from typing import Optional, List

from pydantic import BaseModel, Field

from retk.const import settings, notice


class GetUserRequest(BaseModel):
    email: Optional[str] = Field(max_length=settings.EMAIL_MAX_LENGTH, default=None)
    github: Optional[str] = Field(max_length=50, default=None)
    uid: Optional[str] = Field(max_length=settings.UID_MAX_LENGTH, default=None)
    google: Optional[str] = Field(max_length=settings.EMAIL_MAX_LENGTH, default=None)


class GetManagerDataResponse(BaseModel):
    class Data(BaseModel):
        userCount: int = Field(description="total number of users")
        nodeCount: int = Field(description="total number of nodes")

    requestId: str = Field(max_length=settings.REQUEST_ID_MAX_LENGTH, description="request ID")
    data: Data = Field(description="app data")


class GetUserResponse(BaseModel):
    class User(BaseModel):
        class LastState(BaseModel):
            nodeDisplayMethod: int
            nodeDisplaySortKey: str
            recentSearch: List[str]
            recentCursorSearchSelectedNIds: List[str]

        class Settings(BaseModel):
            language: str
            theme: str
            editorMode: str
            editorFontSize: int
            editorCodeTheme: str
            editorSepRightWidth: float
            editorSideCurrentToolId: str

        id: str
        source: int
        account: str
        nickname: str
        email: str
        avatar: str
        disabled: bool
        createdAt: str
        modifiedAt: str
        usedSpace: int
        type: int
        lastState: LastState
        settings: Settings

    requestId: str = Field(max_length=settings.REQUEST_ID_MAX_LENGTH, description="request ID")
    user: User = Field(description="user info")


class ManagerNoticeDeliveryRequest(BaseModel):
    title: str = Field(
        max_length=settings.MAX_SYSTEM_NOTICE_TITLE_LENGTH,
        description="title"
    )
    content: str = Field(
        max_length=settings.MAX_SYSTEM_NOTICE_CONTENT_LENGTH,
        description="content"
    )
    recipientType: notice.RecipientTypeEnum = Field(..., description="recipient type")
    batchTypeIds: List[str] = Field(
        default_factory=list,
        description="list of user ID if the recipientType is batch"
    )
    publishAt: datetime = Field(..., description="publish time")


class GetSystemNoticesResponse(BaseModel):
    class Notice(BaseModel):
        id: str = Field(description="notice ID")
        title: str = Field(description="title")
        html: str = Field(description="html")
        snippet: str = Field(description="snippet")
        recipientType: notice.RecipientTypeEnum = Field(description="recipient type")
        batchTypeIds: List[str] = Field(description="list of user ID if the recipientType is batch")
        publishAt: datetime = Field(description="publish time")
        scheduled: bool = Field(description="scheduled")

    requestId: str = Field(max_length=settings.REQUEST_ID_MAX_LENGTH, description="request ID")
    notices: List[Notice] = Field(
        description="list of notices"
    )
    total: int = Field(description="total number of notices")



import httpx

from retk import const, config
from retk.controllers import schemas
from retk.controllers.utils import maybe_raise_json_exception, json_exception
from retk.core import account, user, notice, analysis
from retk.models.tps import AuthedUser
from retk.utils import datetime2str


async def __get_then_set_github_user_id(au: AuthedUser, req: schemas.manager.GetUserRequest):
    if req.github.startswith("https://github.com/"):
        req.github = req.github.split("/", 4)[3]
    async with httpx.AsyncClient() as ac:
        url = f"https://api.github.com/users/{req.github}"
        try:
            resp = await ac.get(
                url=url,
                headers={
                    "Accept": "application/vnd.github+json",
                    "Authorization": f"Bearer {config.get_settings().OAUTH_API_TOKEN_GITHUB}",
                    "X-GitHub-Api-Version": "2022-11-28",
                },
                follow_redirects=False,
                timeout=5.
            )
        except (
                httpx.ConnectTimeout,
                httpx.ConnectError,
                httpx.ReadTimeout,
                httpx.HTTPError
        ) as e:
            raise json_exception(
                request_id=au.request_id,
                code=const.CodeEnum.INVALID_PARAMS,
                log_msg=f"get github user info failed, error={e}",
            )
        if resp.status_code != 200:
            raise json_exception(
                request_id=au.request_id,
                code=const.CodeEnum.INVALID_PARAMS,
                log_msg=f"get github user info failed, status_code={resp.status_code}",
            )

        rj = resp.json()
        github_user_id = rj["id"]
        u, code = await user.get_account(
            account=str(github_user_id),
            source=const.UserSourceEnum.GITHUB.value,
            disabled=None,
            exclude_manager=True,
        )
        if code == const.CodeEnum.OK:
            req.uid = u["id"]


async def __get_then_set_google_user_id(req: schemas.manager.GetUserRequest):
    # req.google is a email
    u, code = await user.get_google_account_by_email(
        email=req.google,
        disabled=None,
        exclude_manager=True,
    )
    if code == const.CodeEnum.OK:
        req.uid = u["id"]


async def __check_user_uid(au: AuthedUser, req: schemas.manager.GetUserRequest) -> bool:
    # if all req.dict().values() is None, raise exception
    if all(v is None for v in req.dict().values()):
        raise json_exception(
            request_id=au.request_id,
            code=const.CodeEnum.INVALID_PARAMS,
            log_msg="uid and email and github and google can't be all None",
        )
    if req.github is not None:
        await __get_then_set_github_user_id(au=au, req=req)
    elif req.google is not None:
        await __get_then_set_google_user_id(req=req)
    return req.uid is not None


async def get_manager_data(
        au: AuthedUser,
) -> schemas.manager.GetManagerDataResponse:
    data, code = await analysis.get_marco_data()
    maybe_raise_json_exception(au=au, code=code)
    return schemas.manager.GetManagerDataResponse(
        requestId=au.request_id,
        data=schemas.manager.GetManagerDataResponse.Data(
            userCount=data["user_count"],
            nodeCount=data["node_count"],
        )
    )


async def get_user_info(
        au: AuthedUser,
        req: schemas.manager.GetUserRequest,
) -> schemas.manager.GetUserResponse:
    if await __check_user_uid(au=au, req=req):
        u, code = await user.get(uid=req.uid, disabled=None, exclude_manager=True)
    else:
        u, code = await user.get_by_email(email=req.email, disabled=None, exclude_manager=True)
    maybe_raise_json_exception(au=au, code=code)
    return schemas.manager.GetUserResponse(
        requestId=au.request_id,
        user=schemas.manager.GetUserResponse.User(
            id=u["id"],
            source=u["source"],
            account=u["account"],
            nickname=u["nickname"],
            email=u["email"],
            avatar=u["avatar"],
            disabled=u["disabled"],
            createdAt=datetime2str(u["_id"].generation_time),
            modifiedAt=datetime2str(u["modifiedAt"]),
            usedSpace=u["usedSpace"],
            type=u["type"],
            lastState=schemas.manager.GetUserResponse.User.LastState(
                nodeDisplayMethod=u["lastState"]["nodeDisplayMethod"],
                nodeDisplaySortKey=u["lastState"]["nodeDisplaySortKey"],
                recentSearch=u["lastState"]["recentSearch"],
                recentCursorSearchSelectedNIds=u["lastState"]["recentCursorSearchSelectedNIds"],
            ),
            settings=schemas.manager.GetUserResponse.User.Settings(
                language=u["settings"]["language"],
                theme=u["settings"]["theme"],
                editorMode=u["settings"]["editorMode"],
                editorFontSize=u["settings"]["editorFontSize"],
                editorCodeTheme=u["settings"]["editorCodeTheme"],
                editorSepRightWidth=u["settings"]["editorSepRightWidth"],
                editorSideCurrentToolId=u["settings"]["editorSideCurrentToolId"],
            ),
        ),
    )


async def disable_account(
        au: AuthedUser,
        req: schemas.manager.GetUserRequest,
) -> schemas.RequestIdResponse:
    if await __check_user_uid(au=au, req=req):
        code = await account.manager.disable_by_uid(uid=req.uid)
    else:
        code = await account.manager.disable_by_email(email=req.email)
    maybe_raise_json_exception(au=au, code=code)

    return schemas.RequestIdResponse(
        requestId=au.request_id,
    )


async def enable_account(
        au: AuthedUser,
        req: schemas.manager.GetUserRequest,
) -> schemas.RequestIdResponse:
    if await __check_user_uid(au=au, req=req):
        code = await account.manager.enable_by_uid(uid=req.uid)
    else:
        code = await account.manager.enable_by_email(email=req.email)
    maybe_raise_json_exception(au=au, code=code)
    return schemas.RequestIdResponse(
        requestId=au.request_id,
    )


async def delete_account(
        au: AuthedUser,
        req: schemas.manager.GetUserRequest,
) -> schemas.RequestIdResponse:
    if await __check_user_uid(au=au, req=req):
        await account.manager.delete_by_uid(uid=req.uid)
    else:
        await account.manager.delete_by_email(email=req.email)
    return schemas.RequestIdResponse(
        requestId=au.request_id,
    )


async def post_in_manager_delivery(
        au: AuthedUser,
        req: schemas.manager.ManagerNoticeDeliveryRequest,
) -> schemas.RequestIdResponse:
    _, code = await notice.post_in_manager_delivery(
        au=au,
        title=req.title,
        content=req.content,
        recipient_type=req.recipientType,
        batch_type_ids=req.batchTypeIds,
        publish_at=req.publishAt,
    )
    maybe_raise_json_exception(au=au, code=code)

    return schemas.RequestIdResponse(
        requestId=au.request_id,
    )


async def get_system_notices(
        au: AuthedUser,
        page: int,
        limit: int,
) -> schemas.manager.GetSystemNoticesResponse:
    notices, total = await notice.get_system_notices(
        page=page,
        limit=limit,
    )
    for n in notices:
        n["id"] = str(n["_id"])
        del n["_id"]
    return schemas.manager.GetSystemNoticesResponse(
        requestId=au.request_id,
        notices=notices,
        total=total,
    )



from dataclasses import dataclass


@dataclass
class UserConfig:
    id: int
    max_store_space: int


_mb = 1024 * 1024
_gb = 1024 * _mb


class UserType:
    NORMAL = UserConfig(
        id=0,
        max_store_space=500 * _mb,  # 500MB
    )
    ADMIN = UserConfig(
        id=1,
        max_store_space=100 * _gb,  # 100GB
    )
    MANAGER = UserConfig(
        id=2,
        max_store_space=10 * _gb,  # 10GB
    )

    def id2config(self, _id: int):
        if _id == self.NORMAL.id:
            return self.NORMAL
        elif _id == self.ADMIN.id:
            return self.ADMIN
        elif _id == self.MANAGER.id:
            return self.MANAGER
        else:
            raise ValueError(f"Invalid user type: {_id}")


USER_TYPE = UserType()



from fastapi import Request, APIRouter
from fastapi.params import Annotated, Path
from fastapi.responses import JSONResponse

from retk.controllers import oauth as co
from retk.routes import utils

router = APIRouter(
    prefix="/api/oauth",
    tags=["oauth"],
    responses={404: {"description": "Not found"}},
)


@router.get(
    path="/login/{provider}",
    status_code=200,
    response_model=co.OAuthResponse,
)
@utils.measure_time_spend
async def login_provider(
        provider: str = Annotated[str, Path(title="The provider name", max_length=40)],
) -> co.OAuthResponse:
    return await co.login_provider(provider_name=provider)


@router.get(
    path="/callback/{provider}",
    status_code=200,
)
@utils.measure_time_spend
async def callback_provider(
        request: Request,
        provider: str = Annotated[str, Path(title="The provider name", max_length=40)],
) -> JSONResponse:
    return await co.provider_callback(provider_name=provider, req=request)



import json
from datetime import datetime
from pathlib import Path

from jinja2 import Template

import retk


class DailySummary(retk.Plugin):
    id = "dailySummary"
    name = "Daily Summary"
    version = "0.1.0"
    description = "summary my daily usage"
    author = "morvanzhou"
    icon = Path("_static") / "image" / "logo.svg"

    schedule_timing = retk.scheduler.timing.every_day_at(hour=0, minute=0)

    def __init__(self):
        super().__init__()
        self.data_path = Path(__file__).parent / ".data.json"
        today = datetime.now().strftime("%Y-%m-%d")
        if not self.data_path.exists():
            self.data = [{
                "date": today,
                "word": 0,
                "node": 0,
            }]
            self.data_path.write_text(json.dumps(self.data))
        else:
            self.data = json.loads(self.data_path.read_text())
            if self.data[0]["date"] != today:
                self.data.insert(0, {
                    "date": today,
                    "word": 0,
                    "node": 0,
                })
                self.data = self.data[:120]
                self.write_to_file()

        self.template_home = Template(
            (Path(__file__).parent / "templates" / "home.html").read_text(encoding="utf-8")
        )
        self.template_side = Template(
            (Path(__file__).parent / "templates" / "side.html").read_text(encoding="utf-8")
        )

    def on_node_updated(self, node: retk.tps.Node, old_node: retk.tps.Node) -> None:
        self.data[0]["word"] += len(node["md"]) - len(old_node["md"])
        self.write_to_file()

    def on_node_added(self, node: retk.tps.Node) -> None:
        self.data[0]["word"] += len(node["md"])
        self.data[0]["node"] += 1
        self.write_to_file()

    def on_schedule(self):
        self.data.insert(0, {
            "date": datetime.now().strftime("%Y-%m-%d"),
            "word": 0,
            "node": 0,
        })
        self.data = self.data[:120]
        self.write_to_file()

    def render_plugin_home(self, language: str):
        d = self.data[:7]
        return self.template_home.render(
            name=self.name,
            data=d,
        )

    def render_editor_side(self, uid: str, nid: str, md: str, language: str) -> str:
        d = self.data[:7]
        return self.template_side.render(
            name=self.name,
            data=d,
        )

    def write_to_file(self):
        self.data_path.write_text(json.dumps(self.data, indent=2, ensure_ascii=False))



from . import (  # noqa: F401
    app,
    settings,
    notice,
)
from .languages import LanguageEnum
from .new_user_default import NEW_USER_DEFAULT_NODES, DEFAULT_USER
from .node_display import NodeDisplayMethodEnum, NodeDisplaySortKeyEnum
from .node_types import NodeTypeEnum
from .response_codes import CodeEnum, get_msg_by_code, CODE2STATUS_CODE
from .user_behavior_types import USER_BEHAVIOR_TYPE_MAP, UserBehaviorTypeEnum
from .user_sources import UserSourceEnum
from .user_types import USER_TYPE



"""Microsoft SSO Oauth Helper class"""

from typing import List, Optional

from .base import DiscoveryDocument, OpenID, SSOBase


class MicrosoftSSO(SSOBase):
    """Class providing login using Microsoft OAuth"""

    provider = "microsoft"
    scope = ["openid", "User.Read"]
    version = "v1.0"
    tenant: str = "common"

    def __init__(
            self,
            client_id: str,
            client_secret: str,
            redirect_uri: Optional[str] = None,
            allow_insecure_http: bool = False,
            use_state: bool = False,
            scope: Optional[List[str]] = None,
            tenant: Optional[str] = None,
    ):
        super().__init__(
            client_id=client_id,
            client_secret=client_secret,
            redirect_uri=redirect_uri,
            allow_insecure_http=allow_insecure_http,
            use_state=use_state,
            scope=scope,
        )
        self.tenant = tenant or self.tenant

    async def get_discovery_document(self) -> DiscoveryDocument:
        return {
            "authorization_endpoint": f"https://login.microsoftonline.com/{self.tenant}/oauth2/v2.0/authorize",
            "token_endpoint": f"https://login.microsoftonline.com/{self.tenant}/oauth2/v2.0/token",
            "userinfo_endpoint": f"https://graph.microsoft.com/{self.version}/me",
        }

    @classmethod
    async def openid_from_response(cls, response: dict) -> OpenID:
        return OpenID(
            email=response["mail"],
            display_name=response["displayName"],
            provider=cls.provider,
            id=response.get("id"),
            first_name=response.get("givenName"),
            last_name=response.get("surname"),
        )



# noqa: F401
from retk.models.tps.auth_user import AuthedUser, convert_user_dict_to_authed_user
from retk.models.tps.import_data import ImportData
from retk.models.tps.node import Node, LinkedNode
from retk.models.tps.notice import NoticeManagerDelivery, NoticeSystem
from retk.models.tps.statistic import UserBehavior
from retk.models.tps.user import UserMeta, CODE_THEME_TYPES
from retk.models.tps.user_file import UserFile



import datetime
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import List, Tuple, Sequence, Literal

from retk import const
from retk.models.tps import AuthedUser
from retk.utils import strip_html_tags


@dataclass
class SearchDoc:
    nid: str
    title: str
    body: str

    def __post_init__(self):
        self.title = strip_html_tags(self.title)
        self.body = strip_html_tags(self.body)


@dataclass
class RestoreSearchDoc:
    nid: str
    title: str
    body: str
    createdAt: datetime.datetime
    modifiedAt: datetime.datetime
    disabled: bool
    inTrash: bool

    def __post_init__(self):
        self.title = strip_html_tags(self.title)
        self.body = strip_html_tags(self.body)


@dataclass
class SearchResult:
    nid: str
    score: float
    titleHighlight: str
    bodyHighlights: List[str]


STOPWORDS = sorted(list(set(
    (Path(__file__).parent / "cn_stopwords.txt").read_text(encoding="utf-8").splitlines()
    + (Path(__file__).parent / "baidu_stopwords.txt").read_text(encoding="utf-8").splitlines()
)))


class BaseEngine(ABC):
    hl_tag_name = "em"
    hl_class_name = "match"
    hl_term_prefix = "term"

    @abstractmethod
    async def init(self):
        ...

    @abstractmethod
    async def close(self):
        ...

    @abstractmethod
    async def drop(self):
        ...

    @abstractmethod
    async def add(self, au: AuthedUser, doc: SearchDoc) -> const.CodeEnum:
        ...

    @abstractmethod
    async def update(self, au: AuthedUser, doc: SearchDoc) -> const.CodeEnum:
        ...

    @abstractmethod
    async def to_trash(self, au: AuthedUser, nid: str) -> const.CodeEnum:
        ...

    @abstractmethod
    async def restore_from_trash(self, au: AuthedUser, nid: str) -> const.CodeEnum:
        ...

    @abstractmethod
    async def disable(self, au: AuthedUser, nid: str) -> const.CodeEnum:
        ...

    @abstractmethod
    async def enable(self, au: AuthedUser, nid: str) -> const.CodeEnum:
        ...

    @abstractmethod
    async def delete(self, au: AuthedUser, nid: str) -> const.CodeEnum:
        ...

    @abstractmethod
    async def add_batch(self, au: AuthedUser, docs: List[SearchDoc]) -> const.CodeEnum:
        ...

    @abstractmethod
    async def delete_batch(self, au: AuthedUser, nids: List[str]) -> const.CodeEnum:
        ...

    @abstractmethod
    async def force_delete_all(self, uid: str) -> const.CodeEnum:
        ...

    @abstractmethod
    async def update_batch(self, au: AuthedUser, docs: List[SearchDoc]) -> const.CodeEnum:
        ...

    @abstractmethod
    async def batch_to_trash(self, au: AuthedUser, nids: List[str]) -> const.CodeEnum:
        ...

    @abstractmethod
    async def restore_batch_from_trash(self, au: AuthedUser, nids: List[str]) -> const.CodeEnum:
        ...

    @abstractmethod
    async def _search(
            self,
            au: AuthedUser,
            query: str = "",
            sort_key: str = None,
            reverse: bool = False,
            page: int = 1,
            page_size: int = 10,
            exclude_nids: Sequence[str] = None,
            with_stop_analyzer: bool = False,
    ):
        ...

    @abstractmethod
    async def search(
            self,
            au: AuthedUser,
            query: str = "",
            sort_key: Literal[
                "createdAt", "modifiedAt", "title", "similarity"
            ] = None,
            reverse: bool = False,
            page: int = 1,
            limit: int = 10,
            exclude_nids: Sequence[str] = None,
    ) -> Tuple[List[SearchResult], int]:
        ...

    @abstractmethod
    async def recommend(
            self,
            au: AuthedUser,
            content: str,
            max_return: int = 10,
            exclude_nids: Sequence[str] = None,
    ) -> List[SearchResult]:
        ...

    @abstractmethod
    async def refresh(self):
        ...

    @abstractmethod
    async def count_all(self) -> int:
        ...

    @abstractmethod
    async def batch_restore_docs(self, au: AuthedUser, docs: List[RestoreSearchDoc]) -> const.CodeEnum:
        ...



import os

from starlette.middleware.base import BaseHTTPMiddleware

vue_app_mode = os.getenv("VUE_APP_MODE", "local")

if vue_app_mode not in ["dev", "local"]:
    allow_origins = [
        "https://rethink.run",
        "https://www.rethink.run",
    ]
    csp_local = ""
    cookie_domain = "rethink.run"
    cookie_secure = True
else:
    allow_origins = [
        "*",
    ]
    csp_local = " http://localhost:* http://127.0.0.1:* https://cdn.jsdelivr.net "
    cookie_domain = None
    cookie_secure = False

cookie_samesite = "strict"


class CSPMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request, call_next):
        response = await call_next(request)
        csp_header = "default-src 'self'; " \
                     f"script-src 'self' 'unsafe-inline' 'unsafe-eval' {csp_local} https://www.googletagmanager.com; " \
                     f"style-src 'self' 'unsafe-inline' 'report-sample' {csp_local}; " \
                     "img-src * data: blob:; " \
                     f"connect-src 'self' {csp_local} https://www.google-analytics.com;"
        response.headers["Content-Security-Policy"] = csp_header
        return response


class FrameOptionsMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request, call_next):
        response = await call_next(request)
        response.headers["X-Frame-Options"] = "SAMEORIGIN"
        return response



from . import collection
from . import cursor
from . import database
from . import errors
from . import mongita_client
from . import mongitasync
from . import read_concern
from . import results
from . import write_concern

mongita_client = mongita_client
MongitaClientMemory = mongita_client.MongitaClientMemory
MongitaClientDisk = mongita_client.MongitaClientDisk

database = database
collection = collection
cursor = cursor
errors = errors
results = results
read_concern = read_concern
write_concern = write_concern
mongitasync = mongitasync.mongitasync

ASCENDING = collection.ASCENDING
DESCENDING = collection.DESCENDING

UNIMPLEMENTED = ['ALL', 'CursorType', 'DeleteMany', 'DeleteOne', 'GEO2D', 'GEOHAYSTACK',
                 'GEOSPHERE', 'HASHED', 'IndexModel', 'InsertOne', 'MAX_SUPPORTED_WIRE_VERSION',
                 'MIN_SUPPORTED_WIRE_VERSION', 'MongoClient', 'MongoReplicaSetClient',
                 'OFF', 'ReadPreference', 'ReplaceOne', 'ReturnDocument', 'SLOW_ONLY',
                 'TEXT', 'UpdateMany', 'UpdateOne', 'WriteConcern', 'aggregation',
                 'auth', 'auth_aws', 'bulk', 'change_stream', 'client_options',
                 'client_session', 'collation', 'collection', 'command_cursor',
                 'common', 'compression_support', 'cursor', 'cursor_manager', 'database',
                 'driver_info', 'encryption_options', 'errors', 'get_version_string',
                 'has_c', 'helpers', 'ismaster', 'max_staleness_selectors', 'message',
                 'mongo_client', 'mongo_replica_set_client', 'monitor', 'monitoring',
                 'monotonic', 'network', 'operations', 'periodic_executor', 'pool',
                 'read_concern', 'read_preferences', 'response', 'results', 'saslprep',
                 'server', 'server_description', 'server_selectors', 'server_type',
                 'settings', 'socket_checker', 'son_manipulator', 'srv_resolver',
                 'ssl_context', 'ssl_match_hostname', 'ssl_support', 'thread_util',
                 'topology', 'topology_description', 'uri_parser', 'version',
                 'version_tuple', 'write_concern']

VERSION = "1.2.0"


def __getattr__(attr):
    if attr in UNIMPLEMENTED:
        msg = "%s is not yet implemented. Most PyMongo module-level " \
              "attributes/methods will never be implemented. See the Mongita " \
              "docs." % attr
        raise errors.MongitaNotImplementedError(msg)
    raise AttributeError()



import re

from retk import const

# at least 6 characters, at most 20 characters, at least one letter and one number
VALID_PASSWORD = re.compile(r"^(?=.*[A-Za-z])(?=.*\d).{6,20}$")

EMAIL = re.compile(r'^[.0-9a-zA-Z_]{1,18}@([0-9a-zA-Z-]{1,13}\.){1,}[a-zA-Z]{1,3}$')

MD_CODE = re.compile(r"^```[^\S\r\n]*[a-z]*?\n(.*?)\n```$", re.MULTILINE | re.DOTALL)

OBS_INTERNAL_LINK = re.compile(r"\[\[(.*?)]]")
OBS_INTERNAL_FILE = re.compile(r"!\[\[(.*?)]]")
MD_IMG = re.compile(r"!\[(.*?)]\((?!http)(.*?)\)")

MD_AT_LINK = re.compile(r"\[@.+?]\(([\w/]+?)\)", re.MULTILINE)
NID = re.compile(fr"^[A-Za-z0-9]{{20,{const.settings.NID_MAX_LENGTH}}}$")

ONLY_HTTP_URL = re.compile(r"^https?://\S*$")

MD_FILE_EXT = re.compile(r"^(.+)\.md$")



from typing import List, Dict

from pydantic import BaseModel, NonNegativeInt


class FileUploadProcessResponse(BaseModel):
    requestId: str
    code: NonNegativeInt
    msg: str
    process: int
    type: str
    startAt: str
    running: bool


class VditorFilesResponse(BaseModel):
    class Data(BaseModel):
        errFiles: List[str]
        succMap: Dict[str, str]

    msg: str
    code: NonNegativeInt
    requestId: str
    data: Data


class ImageVditorFetchRequest(BaseModel):
    url: str


class VditorImagesResponse(BaseModel):
    class Data(BaseModel):
        originalURL: str
        url: str

    msg: str
    code: NonNegativeInt
    requestId: str
    data: Data



from pathlib import Path

DOMAIN = "rethink.run"
RETHINK_DIR = Path(__file__).parent.parent
DOT_DATA = ".data"
FRONTEND_DIR = RETHINK_DIR / "dist-local"
MD_MAX_LENGTH = 100_000
REQUEST_ID_MAX_LENGTH = 50
UID_MAX_LENGTH = 30
NID_MAX_LENGTH = 30
FID_MAX_LENGTH = 30
SEARCH_QUERY_MAX_LENGTH = 200
RECOMMEND_CONTENT_MAX_LENGTH = 100
EMAIL_MAX_LENGTH = 100
PASSWORD_MAX_LENGTH = 20
NICKNAME_MAX_LENGTH = 20
IMG_RESIZE_THRESHOLD = 1024 * 1024 * 3  # 3MB
MAX_UPLOAD_FILE_SIZE = 1024 * 1024 * 50  # 50MB
PLUGIN_ID_MAX_LENGTH = 40
MAX_MD_BACKUP_VERSIONS = 10
SEARCH_LIMIT_MAX = 100
MAX_STATISTIC_REMARK_LENGTH = 1000
MAX_SYSTEM_NOTICE_TITLE_LENGTH = 100
MAX_SYSTEM_NOTICE_CONTENT_LENGTH = 5000
MAX_SCHEDULE_JOB_INFO_LEN = 1000

COOKIE_ACCESS_TOKEN = "rethinkAccessToken"
COOKIE_REFRESH_TOKEN = "rethinkRefreshToken"
COOKIE_REFRESH_TOKEN_ID = "rethinkRefreshTokenID"



import json
import os
from pathlib import Path
from typing import Dict, Optional, Union

from bson import ObjectId

from retk import const, utils
from retk._version import __version__


def dump_default_dot_rethink(path: Union[os.PathLike, Path]) -> Dict:
    version = {
        "version": __version__,
        "_id": ObjectId(),
        "id": utils.short_uuid(),
        "email": const.DEFAULT_USER["email"],
        "nickname": const.DEFAULT_USER["nickname"],
        "avatar": const.DEFAULT_USER["avatar"],
        "account": const.DEFAULT_USER["email"],
        "settings": {
            "language": os.getenv("RETHINK_DEFAULT_LANGUAGE", const.LanguageEnum.EN.value),
            "theme": const.app.AppThemeEnum.LIGHT.value,
            "editorMode": const.app.EditorModeEnum.WYSIWYG.value,
            "editorFontSize": 15,
            "editorCodeTheme": const.app.EditorCodeThemeEnum.GITHUB.value,
            "editorSepRightWidth": 200,
            "editorSideCurrentToolId": "",
        }
    }
    with open(path, "w", encoding="utf-8") as f:
        out = version.copy()
        out["_id"] = str(out["_id"])
        json.dump(out, f, indent=2, ensure_ascii=False)
    return version


def load_dot_rethink(path: Union[os.PathLike, Path]) -> Optional[Dict]:
    if not os.path.exists(path):
        return None
    with open(path, "r", encoding="utf-8") as f:
        return json.load(f)



from dataclasses import dataclass
from typing import Union, TYPE_CHECKING

from retk.depend.mongita.collection import Collection

if TYPE_CHECKING:
    from motor.motor_asyncio import AsyncIOMotorCollection


@dataclass
class Collections:
    users: Union[Collection, "AsyncIOMotorCollection"] = None
    nodes: Union[Collection, "AsyncIOMotorCollection"] = None
    import_data: Union[Collection, "AsyncIOMotorCollection"] = None
    user_file: Union[Collection, "AsyncIOMotorCollection"] = None
    notice_manager_delivery: Union[Collection, "AsyncIOMotorCollection"] = None
    notice_system: Union[Collection, "AsyncIOMotorCollection"] = None
    user_behavior: Union[Collection, "AsyncIOMotorCollection"] = None



from typing import Optional, Literal

from fastapi import APIRouter
from fastapi.params import Path, Query
from typing_extensions import Annotated

from retk import const
from retk.controllers import schemas
from retk.controllers.node import node_ops, search
from retk.routes import utils

router = APIRouter(
    prefix="/api/nodes",
    tags=["node"],
    responses={404: {"description": "Not found"}},
)


@router.post(
    path="/",
    status_code=201,
    response_model=schemas.node.NodeResponse,
)
@utils.measure_time_spend
async def post_node(
        au: utils.ANNOTATED_AUTHED_USER,
        req: schemas.node.CreateRequest,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.node.NodeResponse:
    return await node_ops.post_node(
        au=au,
        req=req,
    )


@router.post(
    path="/quick",
    status_code=201,
    response_model=schemas.node.NodeResponse,
)
@utils.measure_time_spend
async def post_quick_node(
        au: utils.ANNOTATED_AUTHED_USER,
        req: schemas.node.CreateRequest,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.node.NodeResponse:
    return await node_ops.post_quick_node(
        au=au,
        req=req,
    )


@router.get(
    path="/",
    status_code=200,
    response_model=schemas.node.NodesSearchResponse,
)
@utils.measure_time_spend
async def get_search_nodes(
        au: utils.ANNOTATED_AUTHED_USER,
        q: str = Query(max_length=const.settings.SEARCH_QUERY_MAX_LENGTH),
        sort: Optional[
            Literal["createdAt", "modifiedAt", "title", "similarity"]
        ] = Query(default="createdAt", max_length=20),
        order: Optional[Literal["asc", "desc"]] = Query(default="asc", alias="ord"),
        page: Optional[int] = Query(default=0, ge=0, alias="p"),
        limit: Optional[int] = Query(default=20, ge=0, le=const.settings.SEARCH_LIMIT_MAX),
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.node.NodesSearchResponse:
    return await search.user_nodes(
        au=au,
        q=q,
        sort=sort,
        order=order,
        page=page,
        limit=limit,
    )


# make sure this is before /{nid} otherwise it will be treated as a nid
@router.get(
    path="/core",
    status_code=200,
    response_model=schemas.node.NodesSearchResponse,
)
@utils.measure_time_spend
async def get_core_nodes(
        au: utils.ANNOTATED_AUTHED_USER,
        p: int = Query(default=0, ge=0, description="page number"),
        limit: int = Query(default=10, ge=0, le=const.settings.SEARCH_LIMIT_MAX),
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.node.NodesSearchResponse:
    return await node_ops.get_core_nodes(
        au=au,
        p=p,
        limit=limit,
    )


@router.get(
    path="/{nid}",
    status_code=200,
    response_model=schemas.node.NodeResponse,
)
@utils.measure_time_spend
async def get_node(
        au: utils.ANNOTATED_AUTHED_USER,
        nid: str = utils.ANNOTATED_NID,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.node.NodeResponse:
    return await node_ops.get_node(
        au=au,
        nid=nid,
    )


@router.get(
    path="/{nid}/at",
    status_code=200,
    response_model=schemas.node.NodesSearchResponse,
)
@utils.measure_time_spend
async def get_at_search(
        au: utils.ANNOTATED_AUTHED_USER,
        nid: str = utils.ANNOTATED_NID,
        q: Optional[str] = Query(default="", max_length=const.settings.SEARCH_QUERY_MAX_LENGTH),
        page: Optional[int] = Query(default=0, ge=0, description="page number"),
        limit: Optional[int] = Query(default=20, ge=0, le=const.settings.SEARCH_LIMIT_MAX),
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.node.NodesSearchResponse:
    return await search.node_at_search(
        au=au,
        nid=nid,
        q=q,
        p=page,
        limit=limit,
    )


@router.get(
    path="/{nid}/recommend",
    status_code=200,
    response_model=schemas.node.NodesSearchResponse,
)
@utils.measure_time_spend
async def get_recommend_nodes(
        au: utils.ANNOTATED_AUTHED_USER,
        nid: str = utils.ANNOTATED_NID,
        content: str = Query(
            max_length=const.settings.SEARCH_QUERY_MAX_LENGTH,
            description="recommend based on this content"
        ),
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.node.NodesSearchResponse:
    return await search.recommend_nodes(
        au=au,
        nid=nid,
        content=content,
    )


@router.get(
    path="/{nid}/history",
    status_code=200,
    response_model=schemas.node.HistEditionsResponse,
)
@utils.measure_time_spend
async def get_hist_editions(
        au: utils.ANNOTATED_AUTHED_USER,
        nid: str = utils.ANNOTATED_NID,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.node.HistEditionsResponse:
    return await node_ops.get_hist_editions(
        au=au,
        nid=nid,
    )


@router.get(
    path="/{nid}/history/{version}/md",
    status_code=200,
    response_model=schemas.node.HistEditionMdResponse,
)
@utils.measure_time_spend
async def get_hist_md(
        au: utils.ANNOTATED_AUTHED_USER,
        nid: str = utils.ANNOTATED_NID,
        version: str = Annotated[str, Path(max_length=30)],
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.node.HistEditionMdResponse:
    return await node_ops.get_hist_edition_md(
        au=au,
        nid=nid,
        version=version,
    )


@router.put(
    path="/{nid}/md",
    status_code=200,
    response_model=schemas.node.NodeResponse,
)
@utils.measure_time_spend
async def put_node_md(
        au: utils.ANNOTATED_AUTHED_USER,
        req: schemas.node.PatchMdRequest,
        nid: str = utils.ANNOTATED_NID,
        referer: Optional[str] = utils.DEPENDS_REFERER,
) -> schemas.node.NodeResponse:
    return await node_ops.update_md(
        au=au,
        req=req,
        nid=nid,
    )



import os
import shutil
import unittest
from unittest.mock import patch, AsyncMock

from retk.depend.mongita import MongitaClientDisk
from retk.models.indexing import not_in_and_create_index


class TestModelIndexing(unittest.IsolatedAsyncioTestCase):
    @patch("retk.depend.mongita.collection.Collection.create_index", new_callable=AsyncMock)
    @patch("retk.depend.mongita.collection.Collection.index_information")
    async def test_not_in_and_create_index(
            self,
            mock_index_information,
            mock_create_index
    ):
        mock_index_information.return_value = {}
        mock_create_index.return_value = None

        async def case(keys, out_index):
            index = await not_in_and_create_index(
                coll=coll,
                index_info={},
                keys=keys,
                unique=True,
            )
            self.assertEqual(out_index, index)

        mongo_path = os.path.join("tmp", "mongo")
        os.makedirs(mongo_path, exist_ok=True)
        mongo = MongitaClientDisk(mongo_path)
        db = mongo["test_db"]
        coll = db["test_coll"]

        await case(["id"], "id_1")
        await case(["account", "source"], "account_1_source_1")
        await case([("uid", 1), ("id", -1)], "uid_1_id_-1")
        await case([("uid", 1), ("fid", -1)], "uid_1_fid_-1")
        await case(["recipientId", "read"], "recipientId_1_read_1")

        with self.assertRaises(ValueError):
            await case([["id"]], "id_1")

        shutil.rmtree(mongo_path)
        await mongo.close()



import datetime
import shutil
import time
import unittest
from copy import deepcopy
from io import BytesIO
from pathlib import Path
from textwrap import dedent
from unittest.mock import patch

import httpx
from PIL import Image
from bson import ObjectId
from bson.tz_util import utc
from fastapi import UploadFile
from starlette.datastructures import Headers

from retk import const, core, config
from retk.controllers.schemas.user import PatchUserRequest
from retk.core.files.importing.async_tasks.utils import update_process
from retk.core.scheduler import tasks
from retk.models import db_ops
from retk.models.client import client
from retk.models.tps import ImportData, AuthedUser, convert_user_dict_to_authed_user
from retk.utils import short_uuid
from . import utils


class LocalModelsTest(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.local")
        # create a fake image
        tmp_path = Path(__file__).parent / "tmp" / "fake.png"
        tmp_path.parent.mkdir(parents=True, exist_ok=True)
        Image.new("RGB", (100, 100)).save(tmp_path)

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.local")
        shutil.rmtree(Path(__file__).parent / "tmp", ignore_errors=True)

    async def asyncSetUp(self) -> None:
        await client.init()
        u, _ = await core.user.get_by_email(email=const.DEFAULT_USER["email"])
        self.au = AuthedUser(
            u=convert_user_dict_to_authed_user(u),
            request_id="xxx",
            language=u["settings"]["language"],
        )

    async def asyncTearDown(self) -> None:
        await client.drop()
        shutil.rmtree(Path(__file__).parent / "tmp" / const.settings.DOT_DATA / "files", ignore_errors=True)
        shutil.rmtree(Path(__file__).parent / "tmp" / const.settings.DOT_DATA / "md", ignore_errors=True)

    async def test_user(self):
        u, code = await core.user.get_by_email(email=const.DEFAULT_USER["email"])
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual("rethink", u["nickname"])
        self.assertIsNotNone(u)

        u, code = await core.user.add(
            account="aaa", source=const.UserSourceEnum.EMAIL.value,
            email="aaa", hashed="bbb", nickname="ccc", avatar="ddd", language=const.LanguageEnum.EN.value)
        self.assertNotEqual("", u["id"])
        self.assertEqual(const.CodeEnum.OK, code)
        _uid = u["id"]

        u, code = await core.user.get(_uid)
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual("ccc", u["nickname"])

        u, code = await core.user.patch(
            au=AuthedUser(
                u=convert_user_dict_to_authed_user(u),
                request_id="xxx",
                language=const.LanguageEnum.EN.value,
            ),
            req=PatchUserRequest(
                nickname="2",
                avatar="3",
                lastState=PatchUserRequest.LastState(
                    nodeDisplayMethod=const.NodeDisplayMethodEnum.LIST.value,
                )
            )
        )
        self.assertEqual(const.CodeEnum.OK, code)

        u, code = await core.user.get(_uid)
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual("bbb", u["hashed"])
        self.assertEqual("2", u["nickname"])
        self.assertEqual("3", u["avatar"])
        self.assertEqual(const.NodeDisplayMethodEnum.LIST.value, u["lastState"]["nodeDisplayMethod"])

        code = await core.account.manager.disable_by_uid(uid=_uid)
        self.assertEqual(const.CodeEnum.OK, code)

        u, code = await core.user.get(uid=_uid)
        self.assertEqual(const.CodeEnum.USER_NOT_EXIST, code)
        self.assertIsNone(u)

        u, code = await core.user.get(uid=_uid, disabled=None)
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertTrue(u["disabled"])

        code = await core.account.manager.enable_by_uid(uid=_uid)
        self.assertEqual(const.CodeEnum.OK, code)

        code = await core.account.manager.disable_by_uid(uid="sdwqdqw")
        self.assertEqual(const.CodeEnum.USER_NOT_EXIST, code)

        await core.account.manager.delete_by_uid(uid=_uid)

    async def test_node(self):
        node, code = await core.node.post(
            au=self.au, md="a" * (const.settings.MD_MAX_LENGTH + 1), type_=const.NodeTypeEnum.MARKDOWN.value
        )
        self.assertEqual(const.CodeEnum.NOTE_EXCEED_MAX_LENGTH, code)
        self.assertIsNone(node)

        u, code = await core.user.get(self.au.u.id)
        self.assertEqual(const.CodeEnum.OK, code)
        used_space = u["usedSpace"]
        node, code = await core.node.post(
            au=self.au, md="[title](/qqq)\nbody", type_=const.NodeTypeEnum.MARKDOWN.value
        )
        self.assertEqual(const.CodeEnum.OK, code)
        u, code = await core.user.get(self.au.u.id)
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(used_space + len(node["md"].encode("utf-8")), u["usedSpace"])
        self.assertEqual("modifiedAt", u["lastState"]["nodeDisplaySortKey"])

        n, code = await core.node.get(au=self.au, nid=node["id"])
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual("title", n["title"])
        self.assertEqual("body", n["snippet"])

        ns, total = await client.search.search(au=self.au)
        self.assertEqual(3, len(ns))
        self.assertEqual(3, total)

        ns, total = await client.search.search(au=self.au, limit=5, page=12, sort_key="createdAt")
        self.assertEqual(0, len(ns))
        self.assertEqual(3, total)

        u, code = await core.user.patch(
            au=self.au,
            req=PatchUserRequest(
                lastState=PatchUserRequest.LastState(
                    nodeDisplaySortKey="createdAt",
                )
            )
        )

        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual("createdAt", u["lastState"]["nodeDisplaySortKey"])
        used_space = u["usedSpace"]
        n, _, code = await core.node.update_md(au=self.au, nid=node["id"], md="title2\nbody2")
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual("title2", n["title"])
        self.assertEqual("title2\nbody2", n["md"])
        self.assertEqual(const.NodeTypeEnum.MARKDOWN.value, n["type"])

        u, code = await core.user.get(self.au.u.id)
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(used_space + (
                len(n["md"].encode("utf-8")) -
                len(node["md"].encode("utf-8"))
        ), u["usedSpace"])

        code = await core.node.disable(au=self.au, nid=node["id"])
        self.assertEqual(const.CodeEnum.OK, code)
        n, code = await core.node.get(au=self.au, nid=node["id"])
        self.assertEqual(const.CodeEnum.NODE_NOT_EXIST, code)

        code = await core.node.to_trash(au=self.au, nid=node["id"])
        self.assertEqual(const.CodeEnum.OK, code)

        code = await core.node.delete(au=self.au, nid=node["id"])
        self.assertEqual(const.CodeEnum.OK, code)
        n, code = await core.node.get(au=self.au, nid=node["id"])
        self.assertIsNone(n)
        self.assertEqual(const.CodeEnum.NODE_NOT_EXIST, code)

        u, code = await core.user.get(self.au.u.id)
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(used_space - len(node["md"].encode("utf-8")), u["usedSpace"])

        nodes, total = await core.node.core_nodes(au=self.au, page=0, limit=10)
        self.assertEqual(2, len(nodes))
        self.assertEqual(2, total)

    async def test_parse_at(self):
        nid1, _ = await core.node.post(
            au=self.au, md="c", type_=const.NodeTypeEnum.MARKDOWN.value,
        )
        nid2, _ = await core.node.post(
            au=self.au, md="我133", type_=const.NodeTypeEnum.MARKDOWN.value,
        )
        md = dedent(f"""title
        fffqw [@c](/n/{nid1['id']})
        fff
        [@我133](/n/{nid2['id']})
        ffq
        """)
        node, code = await core.node.post(
            au=self.au, md=md, type_=const.NodeTypeEnum.MARKDOWN.value
        )
        self.assertEqual(const.CodeEnum.OK, code)
        nodes, total = await client.search.search(
            au=self.au,
            query="",
            sort_key="createdAt",
            reverse=True,
            page=0,
            limit=10,
            exclude_nids=[],
        )
        self.assertEqual(5, len(nodes))
        self.assertEqual(5, total)
        found, total = await client.search.search(au=self.au, query="我")
        self.assertEqual(2, len(found), msg=found)
        self.assertEqual(2, total)

        n, code = await core.node.get(au=self.au, nid=node["id"])
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(2, len(n["toNodeIds"]))

        cache = n["md"]
        n, _, code = await core.node.update_md(au=self.au, nid=node["id"], md=f'{cache}xxxx')
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(cache + "xxxx", n["md"])

        n, code = await core.node.get(au=self.au, nid=nid1['id'])
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(1, len(n["fromNodeIds"]))

        n, _, code = await core.node.update_md(au=self.au, nid=node["id"], md=n["title"])
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(0, len(n["toNodeIds"]))

        n, code = await core.node.get(au=self.au, nid=nid1['id'])
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(0, len(n["fromNodeIds"]))

    async def test_add_set(self):
        node, code = await core.node.post(
            au=self.au, md="title\ntext", type_=const.NodeTypeEnum.MARKDOWN.value
        )
        self.assertEqual(0, len(node["toNodeIds"]))
        self.assertEqual(const.CodeEnum.OK, code)

        res = await db_ops.node_add_to_set(node["id"], "toNodeIds", short_uuid())
        self.assertEqual(1, res.modified_count)
        node, code = await core.node.get(au=self.au, nid=node["id"])
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(1, len(node["toNodeIds"]))

    async def test_cursor_text(self):
        n1, code = await core.node.post(
            au=self.au, md="title\ntext", type_=const.NodeTypeEnum.MARKDOWN.value
        )
        self.assertEqual(const.CodeEnum.OK, code)
        n2, code = await core.node.post(
            au=self.au, md="title2\ntext", type_=const.NodeTypeEnum.MARKDOWN.value
        )
        self.assertEqual(const.CodeEnum.OK, code)

        recom, total = await core.node.search.at(
            au=self.au,
            nid=n2["id"],
            query="text",
            page=0,
            limit=10,
        )
        self.assertEqual(1, len(recom))
        self.assertEqual(1, total)

        recom, total = await core.node.search.at(
            au=self.au,
            nid=n2["id"],  # exclude the second node
            query="",  # return recent nodes only
            page=0,
            limit=10,
        )
        self.assertEqual(2, len(recom))
        self.assertEqual(2, total)

        code = await core.recent.added_at_node(au=self.au, nid=n1["id"], to_nid=n2["id"])
        self.assertEqual(const.CodeEnum.OK, code)

        recom, total = await core.node.search.at(
            au=self.au,
            nid=n1["id"],  # exclude the second node
            query="",
            page=0,
            limit=10,
        )
        self.assertEqual(3, len(recom))
        self.assertEqual(3, total)
        self.assertEqual("Welcome to Rethink", recom[2].title)

    async def test_to_trash(self):
        n1, code = await core.node.post(
            au=self.au, md="title\ntext", type_=const.NodeTypeEnum.MARKDOWN.value
        )
        self.assertEqual(const.CodeEnum.OK, code)
        _, code = await core.node.post(
            au=self.au, md="title2\ntext", type_=const.NodeTypeEnum.MARKDOWN.value
        )
        self.assertEqual(const.CodeEnum.OK, code)

        code = await core.node.to_trash(au=self.au, nid=n1["id"])
        self.assertEqual(const.CodeEnum.OK, code)

        ns, total = await core.node.get_nodes_in_trash(au=self.au, page=0, limit=10)
        self.assertEqual(1, len(ns))
        self.assertEqual(1, total)
        self.assertEqual(n1["id"], ns[0]["id"])

        ns, total = await client.search.search(au=self.au, query="")
        self.assertEqual(3, len(ns))
        self.assertEqual(3, total)

        code = await core.node.restore_from_trash(au=self.au, nid=n1["id"])
        self.assertEqual(const.CodeEnum.OK, code)
        nodes, total = await client.search.search(au=self.au, query="")
        self.assertEqual(4, len(nodes))
        self.assertEqual(4, total)

    async def test_search(self):
        code = await core.recent.put_recent_search(au=self.au, query="a")
        self.assertEqual(const.CodeEnum.OK, code)
        await core.recent.put_recent_search(au=self.au, query="c")
        await core.recent.put_recent_search(au=self.au, query="b")

        doc = await client.coll.users.find_one({"id": self.au.u.id})
        self.assertIsNotNone(doc)
        self.assertEqual(["b", "c", "a"], doc["lastState"]["recentSearch"])

    async def test_batch(self):
        ns = []
        for i in range(10):
            n, code = await core.node.post(
                au=self.au, md=f"title{i}\ntext", type_=const.NodeTypeEnum.MARKDOWN.value
            )
            self.assertEqual(const.CodeEnum.OK, code)
            ns.append(n)

        base_count = 2

        code = await core.node.batch_to_trash(au=self.au, nids=[n["id"] for n in ns[:4]])
        self.assertEqual(const.CodeEnum.OK, code)
        nodes, total = await client.search.search(au=self.au, query="")
        self.assertEqual(6 + base_count, len(nodes))
        self.assertEqual(6 + base_count, total)

        tns, total = await core.node.get_nodes_in_trash(au=self.au, page=0, limit=10)
        self.assertEqual(4, total)
        self.assertEqual(4, len(tns))

        code = await core.node.restore_batch_from_trash(au=self.au, nids=[n["id"] for n in tns[:2]])
        self.assertEqual(const.CodeEnum.OK, code)
        nodes, total = await client.search.search(au=self.au)
        self.assertEqual(8 + base_count, len(nodes))
        self.assertEqual(8 + base_count, total)

        code = await core.node.batch_delete(au=self.au, nids=[n["id"] for n in tns[2:4]])
        self.assertEqual(const.CodeEnum.OK, code)
        tns, total = await core.node.get_nodes_in_trash(au=self.au, page=0, limit=10)
        self.assertEqual(0, total)
        self.assertEqual(0, len(tns))

    async def test_files_upload_process(self):
        now = datetime.datetime.now(tz=utc)
        doc: ImportData = {
            "_id": ObjectId(),
            "uid": "xxx",
            "process": 0,
            "type": "text",
            "startAt": now,
            "running": True,
            "obsidian": {},
            "msg": "",
            "code": const.CodeEnum.OK.value,
        }
        res = await client.coll.import_data.insert_one(doc)
        self.assertTrue(res.acknowledged)

        doc, code = await update_process("xxx", "obsidian", 10)
        self.assertEqual(const.CodeEnum.OK, code)

        doc = await core.files.get_upload_process("xxx1213")
        self.assertIsNone(doc)

        doc = await core.files.get_upload_process("xxx")
        self.assertEqual(10, doc["process"])
        self.assertEqual("obsidian", doc["type"])
        self.assertEqual(now, doc["startAt"])
        self.assertTrue(doc["running"])

        await client.coll.import_data.delete_one({"uid": "xxx"})

    async def test_update_title_and_from_nodes_updates(self):
        n1, code = await core.node.post(
            au=self.au, md="title1\ntext", type_=const.NodeTypeEnum.MARKDOWN.value
        )
        self.assertEqual(const.CodeEnum.OK, code)
        n2, code = await core.node.post(
            au=self.au, md=f"title2\n[@title1](/n/{n1['id']})", type_=const.NodeTypeEnum.MARKDOWN.value
        )
        self.assertEqual(const.CodeEnum.OK, code)

        n1, _, code = await core.node.update_md(au=self.au, nid=n1["id"], md="title1Changed\ntext")
        self.assertEqual(const.CodeEnum.OK, code)
        n2, code = await core.node.get(au=self.au, nid=n2["id"])
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(f"title2\n[@title1Changed](/n/{n1['id']})", n2["md"])

    async def test_upload_image_vditor(self):
        u, code = await core.user.get(self.au.u.id)
        self.assertEqual(const.CodeEnum.OK, code)
        used_space = u["usedSpace"]
        p = Path(__file__).parent / "tmp" / "fake.png"

        image = Image.open(p)
        buf = BytesIO()
        image.save(buf, format="png")
        size = buf.tell()
        img_file = UploadFile(
            buf, filename="fake.png", size=size,
            headers=Headers({"content-type": "image/png"})
        )
        res = await core.files.vditor_upload(au=self.au, files=[img_file])
        self.assertIn("fake.png", res["succMap"])
        self.assertTrue(".png" in res["succMap"]["fake.png"])
        local_file = Path(__file__).parent / "tmp" / const.settings.DOT_DATA / res["succMap"]["fake.png"][1:]
        self.assertTrue(local_file.exists())
        local_file.unlink()
        image.close()
        buf.close()
        await img_file.close()

        u, code = await core.user.get(self.au.u.id)
        self.assertEqual(used_space + size, u["usedSpace"])

    @patch(
        "retk.core.files.upload.httpx.AsyncClient.get",
    )
    async def test_fetch_image_vditor(self, mock_get):
        f = open(Path(__file__).parent / "tmp" / "fake.png", "rb")
        mock_get.return_value = httpx.Response(
            200,
            content=f.read(),
            headers={"content-type": "image/png"}
        )

        u, code = await core.user.get(self.au.u.id)
        used_space = u["usedSpace"]

        url = "https://rethink.run/favicon.png"
        new_url, code = await core.files.fetch_image_vditor(au=self.au, url=url)
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertTrue(new_url.endswith(".png"))
        self.assertTrue(new_url.startswith("/"))
        local_file = Path(__file__).parent / "tmp" / const.settings.DOT_DATA / new_url[1:]
        self.assertTrue(local_file.exists())
        local_file.unlink()

        u, code = await core.user.get(self.au.u.id)
        self.assertEqual(used_space + f.tell(), u["usedSpace"])
        f.close()

    async def test_update_used_space(self):
        u, code = await core.user.get(self.au.u.id)
        base_used_space = u["usedSpace"]
        for delta, value in [
            (100, 100),
            (100, 200),
            (0, 200),
            (-3000, 0),
            (20.1, 20.1),
        ]:
            code = await core.user.update_used_space(uid=self.au.u.id, delta=delta)
            self.assertEqual(const.CodeEnum.OK, code)
            u, code = await core.user.get(self.au.u.id)
            self.assertEqual(const.CodeEnum.OK, code)
            now = u["usedSpace"] - base_used_space
            if now < 0:
                now = 0
                base_used_space = 0
            self.assertAlmostEqual(value, now, msg=f"delta: {delta}, value: {value}")

    async def test_node_version(self):
        node, code = await core.node.post(
            au=self.au, md="[title](/qqq)\nbody", type_=const.NodeTypeEnum.MARKDOWN.value
        )
        self.assertEqual(const.CodeEnum.OK, code)
        md_path = Path(__file__).parent / "tmp" / const.settings.DOT_DATA / "md" / (node["id"] + ".md")
        self.assertTrue(md_path.exists())

        time.sleep(1)

        _, _, code = await core.node.update_md(au=self.au, nid=node["id"], md="title2\nbody2")
        self.assertEqual(const.CodeEnum.OK, code)
        hist_dir = Path(__file__).parent / "tmp" / const.settings.DOT_DATA / "md" / "hist" / node["id"]
        self.assertEqual(1, len(list(hist_dir.glob("*.md"))))

        time.sleep(1)

        _, _, code = await core.node.update_md(au=self.au, nid=node["id"], md="title2\nbody3")
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(2, len(list(hist_dir.glob("*.md"))))

    async def test_md_history(self):
        bi = config.get_settings().MD_BACKUP_INTERVAL
        config.get_settings().MD_BACKUP_INTERVAL = 0.0001
        n1, code = await core.node.post(
            au=self.au, md="title\ntext", type_=const.NodeTypeEnum.MARKDOWN.value
        )
        self.assertEqual(const.CodeEnum.OK, code)
        time.sleep(0.001)

        _, _, code = await core.node.update_md(
            au=self.au, nid=n1["id"], md="title2\ntext",
        )
        self.assertEqual(const.CodeEnum.OK, code)
        time.sleep(0.001)

        _, _, code = await core.node.update_md(
            au=self.au, nid=n1["id"], md="title3\ntext",
        )
        self.assertEqual(const.CodeEnum.OK, code)

        hist, code = await core.node.get_hist_editions(
            au=self.au,
            nid=n1["id"],
        )
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(2, len(hist))

        hist_md, code = await core.node.get_hist_edition_md(
            au=self.au,
            nid=n1["id"],
            version=hist[1],
        )
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual("title2\ntext", hist_md)

        config.get_settings().MD_BACKUP_INTERVAL = bi

    async def test_get_version(self):
        v, code = await core.self_hosted.get_latest_pkg_version()
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(3, len(v))
        for num in v:
            self.assertTrue(isinstance(num, int))

    async def test_system_notice(self):
        au = deepcopy(self.au)
        au.u.type = const.USER_TYPE.MANAGER.id
        publish_at = datetime.datetime.now()
        doc, code = await core.notice.post_in_manager_delivery(
            au=au,
            title="title",
            content="content",
            recipient_type=const.notice.RecipientTypeEnum.ALL.value,
            batch_type_ids=[],
            publish_at=publish_at,
        )
        self.assertEqual(const.CodeEnum.OK, code)

        docs, total = await core.notice.get_system_notices(0, 10)
        self.assertEqual(1, len(docs))
        self.assertEqual(1, total)
        self.assertEqual(doc["_id"], docs[0]["_id"])
        self.assertEqual("title", docs[0]["title"])
        self.assertEqual("content", docs[0]["snippet"])
        self.assertEqual(publish_at.second, docs[0]["publishAt"].second)
        self.assertFalse(docs[0]["scheduled"])

        await tasks.notice.async_deliver_unscheduled_system_notices()
        time.sleep(0.01)
        docs, total = await core.notice.get_system_notices(0, 10)
        self.assertTrue(docs[0]["scheduled"])

    async def test_notice(self):
        au = deepcopy(self.au)
        doc, code = await core.notice.post_in_manager_delivery(
            au=au,
            title="title",
            content="content",
            recipient_type=const.notice.RecipientTypeEnum.ALL.value,
            batch_type_ids=[],
            publish_at=None,
        )
        self.assertEqual(const.CodeEnum.NOT_PERMITTED, code)

        au.u.type = const.USER_TYPE.MANAGER.id
        doc, code = await core.notice.post_in_manager_delivery(
            au=au,
            title="title",
            content="content",
            recipient_type=const.notice.RecipientTypeEnum.ALL.value,
            batch_type_ids=[],
            publish_at=None,
        )
        self.assertEqual(const.CodeEnum.OK, code)

        await tasks.notice.async_deliver_unscheduled_system_notices()

        res = await client.coll.notice_system.find().to_list(None)
        self.assertEqual(1, len(res))
        self.assertEqual(doc["_id"], res[0]["noticeId"])
        self.assertEqual(au.u.id, res[0]["senderId"])
        self.assertFalse(res[0]["read"])

        d = await client.coll.notice_manager_delivery.find_one({"_id": doc["_id"]})
        self.assertIsNotNone(d)
        self.assertTrue(d["scheduled"])

        n, code = await core.notice.get_user_notices(au)
        self.assertEqual(const.CodeEnum.OK, code)
        sn = n["system"]["notices"]
        self.assertEqual(1, len(sn))
        self.assertEqual(1, n["system"]["total"])
        self.assertEqual(str(doc["_id"]), sn[0]["id"])
        self.assertEqual("title", sn[0]["title"])
        self.assertEqual("content", sn[0]["snippet"])
        self.assertLess(datetime.datetime.strptime(sn[0]["publishAt"], '%Y-%m-%dT%H:%M:%SZ'), datetime.datetime.now())
        self.assertFalse(sn[0]["read"])
        self.assertIsNone(sn[0]["readTime"])

        n, code = await core.notice.get_system_notice(au.u.id, notice_id=str(doc["_id"]))
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual("title", n["title"])
        self.assertEqual("<p>content</p>", n["html"])
        self.assertLess(datetime.datetime.strptime(sn[0]["publishAt"], '%Y-%m-%dT%H:%M:%SZ'), datetime.datetime.now())
        self.assertFalse(sn[0]["read"])
        self.assertIsNone(sn[0]["readTime"])

    async def test_mark_read(self):
        au = deepcopy(self.au)
        au.u.type = const.USER_TYPE.MANAGER.id
        for i in range(3):
            _, code = await core.notice.post_in_manager_delivery(
                au=au,
                title=f"title{i}",
                content=f"content{i}",
                recipient_type=const.notice.RecipientTypeEnum.ALL.value,
                batch_type_ids=[],
                publish_at=None,
            )
            self.assertEqual(const.CodeEnum.OK, code)

        await tasks.notice.async_deliver_unscheduled_system_notices()

        au.u.type = const.USER_TYPE.NORMAL.id
        n, code = await core.notice.get_user_notices(au)
        self.assertEqual(const.CodeEnum.OK, code)
        sn = n["system"]["notices"]
        self.assertEqual(3, len(sn))
        self.assertEqual(3, n["system"]["total"])
        for s in sn:
            self.assertFalse(s["read"])
            self.assertIsNone(s["readTime"])

        code = await core.notice.mark_system_notice_read(au.u.id, sn[0]["id"])
        self.assertEqual(const.CodeEnum.OK, code)

        n, code = await core.notice.get_user_notices(au)
        self.assertEqual(const.CodeEnum.OK, code)
        _sn = n["system"]["notices"]
        self.assertEqual(3, len(_sn))
        self.assertEqual(3, n["system"]["total"])
        for i in range(3):
            if _sn[i]["id"] == sn[0]["id"]:
                self.assertTrue(_sn[i]["read"])
                self.assertIsNotNone(_sn[i]["readTime"])
                continue
            self.assertFalse(sn[i]["read"])
            self.assertIsNone(sn[i]["readTime"])

        code = await core.notice.mark_all_system_notice_read(au)
        self.assertEqual(const.CodeEnum.OK, code)

        n, code = await core.notice.get_user_notices(au)
        self.assertEqual(const.CodeEnum.OK, code)
        sn = n["system"]["notices"]
        self.assertEqual(3, len(sn))
        self.assertEqual(3, n["system"]["total"])
        for s in sn:
            self.assertTrue(s["read"])
            self.assertIsNotNone(s["readTime"])



from enum import Enum, unique, IntEnum


@unique
class AppThemeEnum(str, Enum):
    DARK = "dark"
    LIGHT = "light"

    def __str__(self):
        return self.value

    @classmethod
    def is_valid(cls, item: str):
        return item in [v.value for v in cls.__members__.values()]


@unique
class EditorModeEnum(str, Enum):
    WYSIWYG = "wysiwyg"
    IR = "ir"

    def __str__(self):
        return self.value

    @classmethod
    def is_valid(cls, item: str):
        return item in [v.value for v in cls.__members__.values()]


@unique
class EditorCodeThemeEnum(str, Enum):
    DRACULA = "dracula"
    GITHUB = "github"
    EMACS = "emacs"
    TANGO = "tango"

    def __str__(self):
        return self.value

    @classmethod
    def is_valid(cls, item: str):
        return item in [v.value for v in cls.__members__.values()]


@unique
class EditorFontSizeEnum(IntEnum):
    MAX = 30
    MIN = 10

    @classmethod
    def is_valid(cls, item: int):
        return cls.MIN.value <= item <= cls.MAX.value


class FileTypesEnum(Enum):
    IMAGE = {".png", ".jpg", ".jpeg", ".gif", ".webp", ".svg"}
    DOC = {".doc", ".docx", ".xls", ".xlsx", ".ppt", ".pptx"}
    PLAIN = {".txt", ".md", ".csv", ".json", ".xml", ".yaml", ".yml", ".toml"}
    PDF = {".pdf"}
    VIDEO = {".mp4", ".avi", ".mov", ".wmv", ".flv", ".mkv"}
    AUDIO = {".mp3", ".wav", ".wma", ".ogg", ".aac", ".flac"}
    UNKNOWN = "unknown"

    @classmethod
    def get_type(cls, ext: str):
        for t in cls.__members__.values():
            if ext in t.value:
                return t
        return cls.UNKNOWN


@unique
class ValidUploadedFilePrefixEnum(Enum):
    # image/*,video/*,audio/*,application/pdf,text/plain,text/markdown
    IMAGE = "image/"
    VIDEO = "video/"
    AUDIO = "audio/"
    TEXT = "text/"
    PDF = "application/pdf"
    DOC = "application/msword"
    DOCX = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    XLS = "application/vnd.ms-excel"
    XLSX = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
    PPT = "application/vnd.ms-powerpoint"
    PPTX = "application/vnd.openxmlformats-officedocument.presentationml.presentation"

    def __str__(self):
        return self.value

    @classmethod
    def is_valid(cls, item: str):
        for pre in [v.value for v in cls.__members__.values()]:
            if pre.endswith("/") and item.startswith(pre):
                return True
            elif pre == item:
                return True
        return False



import datetime
from functools import lru_cache
from typing import Optional

import bcrypt
from cryptography.hazmat.backends import default_backend as crypto_default_backend
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from pydantic import Field, DirectoryPath
from pydantic_settings import BaseSettings, SettingsConfigDict

from retk.logger import logger


class Settings(BaseSettings):
    ONE_USER: bool = Field(default=1, env='ONE_USER')
    RETHINK_SERVER_DEBUG: bool = Field(default=False, env='RETHINK_SERVER_DEBUG')
    VERIFY_REFERER: bool = Field(default=False, env='VERIFY_REFERER')
    PLUGINS: bool = Field(default=False, env='PLUGINS')
    RETHINK_LOCAL_STORAGE_PATH: Optional[DirectoryPath] = Field(env='RETHINK_LOCAL_STORAGE_PATH', default=None)
    DB_NAME: str = Field(env='DB_NAME', default="")
    DB_USER: str = Field(env='DB_USER', default="")
    DB_PASSWORD: str = Field(env='DB_PASSWORD', default="")
    DB_HOST: str = Field(env='DB_HOST', default="")
    DB_PORT: int = Field(env='DB_PORT', default=-1)
    DB_SALT: str = Field(env='BD_SALT', default="")
    ES_USER: str = Field(env='ES_USER', default="")
    ES_PASSWORD: str = Field(env='ES_PASSWORD', default="")
    ES_HOSTS: str = Field(env='ES_HOSTS', default="")
    ES_INDEX_ALIAS: str = Field(env='ES_INDEX_ALIAS', default="")
    CAPTCHA_SALT: str = Field(env='CAPTCHA_SALT', default="")
    RETHINK_EMAIL: str = Field(env='RETHINK_EMAIL', default="")
    RETHINK_EMAIL_PASSWORD: str = Field(env='RETHINK_EMAIL_PASSWORD', default="")
    JWT_KEY: bytes = Field(env='JWT_KEY', default=b"")
    JWT_KEY_PUB: bytes = Field(env='JWT_KEY_PUB', default=b"")
    JWT_REFRESH_EXPIRED_DAYS: int = Field(default=1, env='JWT_REFRESH_EXPIRED_DAYS')
    JWT_ACCESS_EXPIRED_MINS: int = Field(default=5, env='JWT_ACCESS_EXPIRED_MINS')
    REFRESH_TOKEN_EXPIRE_DELTA: datetime.timedelta = Field(default=datetime.timedelta(days=1))
    ACCESS_TOKEN_EXPIRE_DELTA: datetime.timedelta = Field(default=datetime.timedelta(minutes=15))
    OAUTH_REDIRECT_URL: str = Field(env='OAUTH_REDIRECT_URL', default="")
    OAUTH_CLIENT_ID_GITHUB: str = Field(env='OAUTH_CLIENT_ID_GITHUB', default="")
    OAUTH_CLIENT_SEC_GITHUB: str = Field(env='OAUTH_CLIENT_SEC_GITHUB', default="")
    OAUTH_API_TOKEN_GITHUB: str = Field(env='OAUTH_API_TOKEN_GITHUB', default="")
    OAUTH_CLIENT_ID_QQ: str = Field(env='OAUTH_CLIENT_ID_QQ', default="")
    OAUTH_CLIENT_SEC_QQ: str = Field(env='OAUTH_CLIENT_SEC_QQ', default="")
    OAUTH_CLIENT_ID_FACEBOOK: str = Field(env='OAUTH_CLIENT_ID_QQ', default="")
    OAUTH_CLIENT_SEC_FACEBOOK: str = Field(env='OAUTH_CLIENT_SEC_QQ', default="")
    OAUTH_CLIENT_ID_GOOGLE: str = Field(env='OAUTH_CLIENT_ID_GOOGLE', default="")
    OAUTH_CLIENT_SEC_GOOGLE: str = Field(env='OAUTH_CLIENT_SEC_GOOGLE', default="")
    COS_SECRET_ID: str = Field(env='COS_SECRET_ID', default="")
    COS_SECRET_KEY: str = Field(env="COS_SECRET_KEY", default="")
    COS_REGION: str = Field(env="COS_REGION", default="")
    COS_BUCKET_NAME: str = Field(env="COS_BUCKET_NAME", default="")
    MD_BACKUP_INTERVAL: int = Field(env="MD_BACKUP_INTERVAL", default=60 * 5)  # 5 minutes

    model_config = SettingsConfigDict(
        env_file=".env",
        env_file_encoding='utf-8',
        case_sensitive=True,
    )

    def __init__(self):
        super().__init__()
        if self.RETHINK_SERVER_DEBUG:
            logger.setLevel("DEBUG")

        if self.RETHINK_LOCAL_STORAGE_PATH is None and self.DB_HOST == "":
            raise ValueError("RETHINK_LOCAL_STORAGE_PATH and DB_HOST cannot be empty at the same time")
        if self.DB_SALT == "":
            self.DB_SALT = bcrypt.gensalt(4).decode("utf-8")
        if self.JWT_KEY == b"" or self.JWT_KEY_PUB == b"":
            key = rsa.generate_private_key(
                backend=crypto_default_backend(),
                public_exponent=65537,
                key_size=4096
            )
            self.JWT_KEY = key.private_bytes(
                crypto_serialization.Encoding.PEM,
                crypto_serialization.PrivateFormat.PKCS8,
                crypto_serialization.NoEncryption()
            )

            self.JWT_KEY_PUB = key.public_key().public_bytes(
                crypto_serialization.Encoding.OpenSSH,
                crypto_serialization.PublicFormat.OpenSSH
            )

        self.REFRESH_TOKEN_EXPIRE_DELTA = datetime.timedelta(days=self.JWT_REFRESH_EXPIRED_DAYS)
        self.ACCESS_TOKEN_EXPIRE_DELTA = datetime.timedelta(minutes=self.JWT_ACCESS_EXPIRED_MINS)


@lru_cache()
def get_settings() -> Settings:
    return Settings()


def is_local_db() -> bool:
    return get_settings().RETHINK_LOCAL_STORAGE_PATH is not None



from typing import Callable

from retk.const import CodeEnum
from retk.models.tps import AuthedUser
from .base import event_plugin_map


def on_node_added(func: Callable):
    async def wrapper(*args, **kwargs):
        data, code = await func(*args, **kwargs)
        if code != CodeEnum.OK:
            return data, code
        for inst in event_plugin_map["on_node_added"]:
            # execute the class method
            inst.on_node_added(node=data)
        return data, code

    return wrapper


def on_node_updated(func: Callable):
    async def wrapper(
            au: AuthedUser,
            nid: str,
            md: str,
            *args, **kwargs,
    ):
        data, old_data, code = await func(
            au=au,
            nid=nid,
            md=md,
            *args, **kwargs
        )
        if code != CodeEnum.OK:
            return data, code
        for inst in event_plugin_map["on_node_updated"]:
            # execute the class method
            inst.on_node_updated(node=data, old_node=old_data)
        return data, old_data, code

    return wrapper


def before_node_updated(func: Callable):
    async def wrapper(
            au: AuthedUser,
            nid: str,
            md: str,
            *args, **kwargs,
    ):
        data = {"md": md}
        for inst in event_plugin_map["before_node_updated"]:
            # execute the class method
            inst.before_node_updated(
                uid=au.u.id,
                nid=nid,
                data=data,
            )
        return await func(
            au=au,
            nid=nid,
            md=data["md"],
            *args, **kwargs
        )

    return wrapper



from collections import OrderedDict as _OrderedDict
from dataclasses import dataclass
from datetime import datetime, timedelta
from functools import wraps
from typing import Callable, Optional, Tuple, Dict, Any, List, OrderedDict

from apscheduler.schedulers.background import BackgroundScheduler
from bson.tz_util import utc

from retk import const
from retk.const.settings import MAX_SCHEDULE_JOB_INFO_LEN
from . import tasks

"""
- BlockingScheduler:
 use when the scheduler is the only thing running in your process

- BackgroundScheduler:
 use when you’re not using any of the frameworks below, 
 and want the scheduler to run in the background inside your application

- AsyncIOScheduler:
 use if your application uses the asyncio module

- GeventScheduler:
 use if your application uses gevent

- TornadoScheduler:
 use if you’re building a Tornado application

- TwistedScheduler: 
 use if you’re building a Twisted application

- QtScheduler:
 use if you’re building a Qt application
"""


@dataclass
class JobInfo:
    id: str
    type: str
    args: Tuple
    kwargs: Dict[str, Any]
    executed_at: Optional[datetime] = None
    finished_at: Optional[datetime] = None
    finished_return: Optional[str] = None
    created_at: datetime = None

    def __post_init__(self):
        self.created_at = datetime.now(tz=utc)

    def executing_time(self) -> timedelta:
        if self.executed_at is None:
            return timedelta(seconds=0)
        if self.finished_at is None:
            return datetime.now() - self.executed_at
        return self.finished_at - self.executed_at

    def is_failed(self) -> bool:
        return self.finished_at is not None and self.executed_at is None


# a separate thread
__scheduler: Optional[BackgroundScheduler] = None
__jobs_info: OrderedDict[str, JobInfo] = _OrderedDict()


def __wrap_func(func: Callable, job_info: JobInfo) -> Optional[Callable]:
    if job_info.id in __jobs_info or __scheduler.get_job(job_info.id) is not None:
        raise KeyError(f"job id {job_info.id} already exists")

    __jobs_info[job_info.id] = job_info
    if len(__jobs_info) > MAX_SCHEDULE_JOB_INFO_LEN:
        # dequeue the oldest job
        __jobs_info.popitem(last=False)

    @wraps(func)
    def wrapper(*args, **kwargs):
        job_info.executed_at = datetime.now(tz=utc)
        res = func(*args, **kwargs)
        job_info.finished_at = datetime.now(tz=utc)
        job_info.finished_return = res

    return wrapper


def init_tasks():
    # check unscheduled system notices every minute
    run_every_at(
        job_id="deliver_unscheduled_system_notices",
        func=tasks.notice.deliver_unscheduled_system_notices,
        second=0,
    )
    return


def get_jobs() -> List[JobInfo]:
    # from oldest to newest
    return list(__jobs_info.values())


def get_job(job_id: str) -> Optional[JobInfo]:
    return __jobs_info.get(job_id)


def clear_jobs() -> None:
    __jobs_info.clear()


def start():
    global __scheduler
    __scheduler = BackgroundScheduler()
    __scheduler.start()


def stop():
    __scheduler.remove_all_jobs()
    __scheduler.shutdown()


def _get_default(args, kwargs):
    if args is None:
        args = ()
    if kwargs is None:
        kwargs = {}
    return args, kwargs


def run_once_at(
        job_id: str,
        func: Callable,
        time: datetime,
        args: Optional[Tuple] = None,
        kwargs: Optional[Dict[str, Any]] = None,
) -> Tuple[JobInfo, const.CodeEnum]:
    args, kwargs = _get_default(args, kwargs)
    ji = JobInfo(
        id=job_id,
        type="date",
        args=args,
        kwargs=kwargs,
    )
    _func = __wrap_func(func=func, job_info=ji)
    if _func is None:
        return ji, const.CodeEnum.INVALID_SCHEDULE_JOB_ID
    __scheduler.add_job(
        id=job_id,
        func=_func,
        trigger="date",
        run_date=time,
        args=args,
        kwargs=kwargs,
    )
    return ji, const.CodeEnum.OK


def run_once_now(
        job_id: str,
        func: Callable,
        args: Optional[Tuple] = None,
        kwargs: Optional[Dict[str, Any]] = None,
) -> Tuple[JobInfo, const.CodeEnum]:
    return run_once_at(job_id=job_id, func=func, time=datetime.now(tz=utc), args=args, kwargs=kwargs)


def run_once_after(
        job_id: str,
        func: Callable,
        second: float,
        args: Optional[Tuple] = None,
        kwargs: Optional[Dict[str, Any]] = None,
) -> Tuple[JobInfo, const.CodeEnum]:
    return run_once_at(
        job_id=job_id,
        func=func,
        time=datetime.now(tz=utc) + timedelta(seconds=second),
        args=args,
        kwargs=kwargs,
    )


def run_every_at(
        job_id: str,
        func: Callable,
        second: Optional[int] = None,
        minute: Optional[int] = None,
        hour: Optional[int] = None,
        day: Optional[int] = None,
        week: Optional[int] = None,
        month: Optional[int] = None,
        day_of_week: Optional[int] = None,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        args: Optional[Tuple] = None,
        kwargs: Optional[Dict[str, Any]] = None,
) -> Tuple[JobInfo, const.CodeEnum]:
    if second is not None:
        second = int(second)
        if second > 59 or second < 0:
            raise ValueError("second must be in range 0-59")
    if minute is not None:
        minute = int(minute)
        if minute > 59 or minute < 0:
            raise ValueError("minute must be in range 0-59")
    if hour is not None:
        hour = int(hour)
        if hour > 23 or hour < 0:
            raise ValueError("hour must be in range 0-23")
    if day is not None:
        day = int(day)
        if day > 31 or day < 1:
            raise ValueError("day must be in range 1-31")
    if week is not None:
        week = int(week)
        if week > 53 or week < 1:
            raise ValueError("week must be in range 1-53")
    if month is not None:
        month = int(month)
        if month > 12 or month < 1:
            raise ValueError("month must be in range 1-12")
    if day_of_week is not None:
        day_of_week = int(day_of_week)
        if day_of_week > 6 or day_of_week < 0:
            raise ValueError("day_of_week must be in range 0-6")

    args, kwargs = _get_default(args, kwargs)
    ji = JobInfo(
        id=job_id,
        type="cron",
        args=args,
        kwargs=kwargs,
    )
    _func = __wrap_func(func=func, job_info=ji)
    if _func is None:
        return ji, const.CodeEnum.INVALID_SCHEDULE_JOB_ID
    __scheduler.add_job(
        id=job_id,
        func=_func,
        trigger="cron",
        second=second,
        minute=minute,
        hour=hour,
        day=day,
        month=month,
        day_of_week=day_of_week,
        start_date=start_date,
        end_date=end_date,
        args=args,
        kwargs=kwargs,
    )
    return ji, const.CodeEnum.OK



from datetime import datetime
from typing import Dict, TypedDict

from bson import ObjectId


class ImportData(TypedDict):
    _id: ObjectId
    uid: str
    process: int
    type: str
    startAt: datetime
    running: bool
    msg: str
    code: int
    obsidian: Dict[str, str]  # filename: nid



import base64
import hashlib
import os
from typing import Optional, Tuple

import bcrypt

from retk import config, const, regex
from retk.core import user, node
from retk.logger import logger
from retk.models import tps
from retk.models.client import client


async def is_right_password(email: str, hashed: str, password: str) -> bool:
    if config.get_settings().ONE_USER:
        pw = os.getenv("RETHINK_SERVER_PASSWORD", None)
        if pw is not None:
            return pw == password
        return True
    base_pw = _base_password(password=password, email=email)
    match = bcrypt.checkpw(base_pw, hashed.encode("utf-8"))
    return match


def _base_password(password: str, email: str) -> bytes:
    # update hash strategy
    s = f"{password}&&{config.get_settings().DB_SALT}$${email}"
    logger.debug(f"hashing: {s}")
    return base64.b64encode(
        hashlib.sha256(s.encode("utf-8")).digest()
    )


def hash_password(password: str, email: str) -> str:
    bpw = _base_password(password=password, email=email)
    salt = bcrypt.gensalt()
    return bcrypt.hashpw(bpw, salt=salt).decode("utf-8")


def login_by_email_pwd(email: str, password: str) -> const.CodeEnum:
    if config.get_settings().ONE_USER:
        logger.warning("on ONE_USER mode, user registration will be skipped")
        return const.CodeEnum.ONE_USER_MODE
    if regex.EMAIL.match(email) is None:
        return const.CodeEnum.INVALID_EMAIL
    if regex.VALID_PASSWORD.match(password) is None:
        return const.CodeEnum.INVALID_PASSWORD
    return const.CodeEnum.OK


async def signup(
        email: str,
        password: str,
        language: str = const.LanguageEnum.EN.value,
) -> Tuple[Optional[tps.UserMeta], const.CodeEnum]:
    code = login_by_email_pwd(email=email, password=password)
    if code != const.CodeEnum.OK:
        return None, code
    u, code = await user.get_by_email(email=email)
    if code == const.CodeEnum.OK or u is not None:
        return None, const.CodeEnum.USER_EXIST

    u, code = await user.add(
        account=email,
        source=const.UserSourceEnum.EMAIL.value,
        email=email,
        hashed=hash_password(password=password, email=email),
        nickname=email.split("@")[0],
        avatar="",
        language=language,
    )
    if code != const.CodeEnum.OK:
        return None, code

    code = await node.new_user_add_default_nodes(uid=u["id"], language=language)
    return u, code


async def __delete_post_process(uid: str):
    await client.coll.nodes.delete_many({"uid": uid})
    await client.coll.user_file.delete_many({"uid": uid})
    await client.coll.import_data.delete_many({"uid": uid})
    await client.coll.notice_system.delete_many({"recipientId": uid})
    await client.coll.user_behavior.delete_many({"uid": uid})
    await client.search.force_delete_all(uid=uid)


async def delete_by_uid(uid: str):
    res = await client.coll.users.delete_one({"id": uid})
    if res.deleted_count != 1:
        return
    await __delete_post_process(uid=uid)


async def delete_by_email(email: str):
    u, code = await user.get_by_email(email=email, exclude_manager=True)
    if code != const.CodeEnum.OK or u is None:
        return
    return await delete_by_uid(uid=u["id"])


async def __disable_enable(condition: dict, disable: bool) -> const.CodeEnum:
    res = await client.coll.users.update_one(
        condition,
        {"$set": {"disabled": disable}}
    )
    return const.CodeEnum.OK if res.matched_count == 1 else const.CodeEnum.USER_NOT_EXIST


async def disable_by_uid(uid: str) -> const.CodeEnum:
    return await __disable_enable({"id": uid}, disable=True)


async def disable_by_email(email: str) -> const.CodeEnum:
    return await __disable_enable({"source": const.UserSourceEnum.EMAIL.value, "account": email}, disable=True)


async def enable_by_uid(uid: str) -> const.CodeEnum:
    return await __disable_enable({"id": uid}, disable=False)


async def enable_by_email(email: str) -> const.CodeEnum:
    return await __disable_enable({"source": const.UserSourceEnum.EMAIL.value, "account": email}, disable=False)



import unittest

from retk import const, core
from retk.models.client import client
from retk.models.tps import AuthedUser, convert_user_dict_to_authed_user
from . import utils


class DataRestoreTest(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.local")

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.local")

    async def asyncSetUp(self) -> None:
        await client.init()

    async def asyncTearDown(self) -> None:
        await client.drop()

    async def test_restore_search(self):
        u, _ = await core.user.get_by_email(email=const.DEFAULT_USER["email"])
        self.au = AuthedUser(
            u=convert_user_dict_to_authed_user(u),
            request_id="test",
            language=const.LanguageEnum.EN.value,
        )
        base_count = 2
        nids = []
        for i in range(20):
            n, code = await core.node.post(
                au=self.au,
                md=f"title{i}\ntext{i}",
                type_=const.NodeTypeEnum.MARKDOWN.value,
            )
            nids.append(n["id"])
            self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(20 + base_count, await client.search.count_all())

        code = await core.node.batch_to_trash(au=self.au, nids=nids[:10])
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(20 + base_count, await client.search.count_all())

        code = await client.search.delete_batch(
            au=self.au,
            nids=nids[:10],
        )
        self.assertEqual(const.CodeEnum.OK, code)
        self.assertEqual(10 + base_count, await client.search.count_all())

        await client.init()
        self.assertEqual(20 + base_count, await client.search.count_all())



from typing import Literal, Optional

from retk import core, const
from retk.controllers import schemas
from retk.models.tps import AuthedUser


async def user_nodes(
        au: AuthedUser,
        q: str,
        sort: Literal["createdAt", "modifiedAt", "title", "similarity"],
        order: Literal["asc", "desc"],
        page: int,
        limit: int,
) -> schemas.node.NodesSearchResponse:
    nodes, total = await core.node.search.user_nodes(
        au=au,
        query=q,
        sort_key=sort,
        reverse=order == "desc",
        page=page,
        limit=limit,
        exclude_nids=[],
    )

    if q != "":
        await core.statistic.add_user_behavior(
            uid=au.u.id,
            type_=const.UserBehaviorTypeEnum.SEARCH_GLOBAL,
            remark=q,
        )
    return schemas.node.NodesSearchResponse(
        requestId=au.request_id,
        data=schemas.node.NodesSearchResponse.Data(
            nodes=nodes,
            total=total,
        )
    )


async def node_at_search(
        au: AuthedUser,
        nid: str,
        q: Optional[str],
        p: int,
        limit: int,
) -> schemas.node.NodesSearchResponse:
    nodes, total = await core.node.search.at(
        au=au,
        nid=nid,
        query=q,
        page=p,
        limit=limit,
    )
    if q != "":
        await core.statistic.add_user_behavior(
            uid=au.u.id,
            type_=const.UserBehaviorTypeEnum.SEARCH_AT,
            remark=q,
        )
    return schemas.node.NodesSearchResponse(
        requestId=au.request_id,
        data=schemas.node.NodesSearchResponse.Data(
            nodes=nodes,
            total=total,
        )
    )


async def recommend_nodes(
        au: AuthedUser,
        nid: str,
        content: str,
) -> schemas.node.NodesSearchResponse:
    max_return = 5
    nodes = await core.node.search.recommend(
        au=au,
        content=content,
        max_return=max_return,
        exclude_nids=[nid],
    )
    return schemas.node.NodesSearchResponse(
        requestId=au.request_id,
        data=schemas.node.NodesSearchResponse.Data(
            nodes=nodes,
            total=len(nodes),
        )
    )



from datetime import datetime
from typing import List, Optional, TypedDict

from bson import ObjectId


# not used, just for reference
class LinkedNode(TypedDict):
    _id: ObjectId
    id: str
    md: str
    title: str
    type: int  # const.NodeType.MARKDOWN.value
    disabled: bool
    modifiedAt: datetime


class Node(TypedDict):
    _id: ObjectId
    id: str
    uid: str
    md: str
    title: str
    snippet: str
    type: int  # const.NodeType.MARKDOWN.value
    disabled: bool
    inTrash: bool
    modifiedAt: datetime
    inTrashAt: Optional[datetime]
    fromNodeIds: List[str]
    toNodeIds: List[str]
    history: List[str]



import json
from collections import OrderedDict
from datetime import datetime
from pathlib import Path
from typing import Any

from jinja2 import Template

import retk


class Favorites(retk.Plugin):
    id = "favorites"
    name = "Favorites"
    version = "0.1.0"
    description = "favorites thoughts"
    author = "morvanzhou"
    icon = Path("_static") / "logo.svg"

    def __init__(self):
        super().__init__()
        self.data_path = Path(__file__).parent / ".data.json"
        if not self.data_path.exists():
            self.data = OrderedDict()
            self.data_path.write_text(json.dumps(self.data))
        else:
            self.data = OrderedDict(json.loads(self.data_path.read_text()))

        self.template_home = Template(
            (Path(__file__).parent / "templates" / "home.html").read_text(encoding="utf-8")
        )
        self.template_side = Template(
            (Path(__file__).parent / "templates" / "side.html").read_text(encoding="utf-8")
        )

    def render_plugin_home(self, language: str) -> str:
        return self.template_home.render(
            add_btn_name="Add to Favorites" if language == "en" else "添加到收藏夹",
            h1=self.name,
            items=list(self.data.values()),
            plugin_id=self.id,
            call_url=self.api_call_url,
        )

    def render_editor_side(self, uid: str, nid: str, md: str, language: str) -> str:
        return self.template_side.render(
            add_btn_name="Add to Favorites" if language == "en" else "添加到收藏夹",
            h1=self.name,
            items=list(self.data.values()),
            nid=nid,
            title=md.strip().split("\n", 1)[0],
            plugin_id=self.id,
            call_url=self.api_call_url,
        )

    def handle_api_call(self, method: str, data: Any) -> retk.PluginAPICallReturn:
        if method == "add":
            if data["nid"] in self.data:
                return retk.PluginAPICallReturn(
                    success=False,
                    message=f"nid='{data['nid']}' already in favorites",
                    data=None
                )
            self.update_file(add=data)
        elif method == "remove":
            if data["nid"] not in self.data:
                return retk.PluginAPICallReturn(
                    success=False,
                    message=f"nid='{data['nid']}' not in favorites",
                    data=None
                )
            self.update_file(remove_nid=data["nid"])
        else:
            return retk.PluginAPICallReturn(
                success=False,
                message=f"method='{method}' is not allowed",
                data=None
            )
        return retk.PluginAPICallReturn(
            success=True,
            message="success",
            data=None
        )

    def update_file(self, add: dict = None, remove_nid: str = None):
        if add is not None:
            add["time"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            self.data[add["nid"]] = add
        if remove_nid is not None:
            self.data.pop(remove_nid)
        self.data_path.write_text(json.dumps(self.data, indent=2, ensure_ascii=False))



from typing import List

from fastapi import UploadFile

from retk import const
from retk.core.files.saver import saver, File


async def save_editor_upload_files(
        uid: str,
        files: List[UploadFile],
) -> dict:
    res = {
        "errFiles": [],
        "succMap": {},
        "code": const.CodeEnum.OK,
    }
    for file in files:
        filename = file.filename
        # validate MIME image type
        if not const.app.ValidUploadedFilePrefixEnum.is_valid(file.content_type):
            res["errFiles"].append(filename)
            res["code"] = const.CodeEnum.INVALID_FILE_TYPE
            continue

        # validate file type
        sep = filename.rsplit(".", 1)
        ext = f".{sep[-1]}" if len(sep) > 1 else ""
        if const.app.FileTypesEnum.get_type(ext) == const.app.FileTypesEnum.UNKNOWN:
            res["errFiles"].append(filename)
            res["code"] = const.CodeEnum.INVALID_FILE_TYPE
            continue

        # validate file size
        if file.size > const.settings.MAX_UPLOAD_FILE_SIZE:
            res["errFiles"].append(filename)
            res["code"] = const.CodeEnum.TOO_LARGE_FILE
            continue
        f = File(
            filename=filename,
            data=file.file,
        )
        if f.is_unknown_type():
            res["errFiles"].append(filename)
            res["code"] = const.CodeEnum.INVALID_FILE_TYPE
            continue

        url = await saver.save(
            uid=uid,
            file=f
        )
        if url == "":
            res["errFiles"].append(filename)
            res["code"] = const.CodeEnum.FILE_OPEN_ERROR
            continue
        res["succMap"][filename] = url
    return res



import abc


class Engine(abc.ABC):
    @abc.abstractmethod
    def put_doc(self, location, doc, generation=None):
        """
        :param location Location: Location obj
        :param doc dict: document as a dict
        :param generation int: document generation.
        :rtype: bool

        Given a location, upload a doc to that location.
        If a generation was provided, check that the generation is newer
        than the storage generation before uploading
        Returns boolean indicating success
        """

    @abc.abstractmethod
    def get_doc(self, location):
        """
        :param location Location: Location obj
        :rtype: dict

        Given a location, download a doc
        If the doc doesn't exist, return None
        """

    @abc.abstractmethod
    def doc_exists(self, location):
        """
        :param location Location: Location obj
        :rtype: bool

        Given a location, return True if the object exists and False otherwise
        """

    @abc.abstractmethod
    def list_ids(self, collection_location, limit=None):
        """
        :param collection_location Location: Location obj|
        :param limit int:
        :rtype: list[str]

        Given a Location without an _id, return a list of _ids.
        """

    @abc.abstractmethod
    def delete_doc(self, location):
        """
        :param location Location: Location obj
        :rtype: bool

        Given a location, delete the document. Return True if found and deleted.
        """

    @abc.abstractmethod
    def delete_dir(self, location):
        """
        :param location Location: Location obj
        :rtype: bool

        Given a location, delete the directory. Return True if found and deleted.
        """

    @abc.abstractmethod
    def create_path(self, location):
        """
        :param location Location: Location obj
        :rtype: bool

        Prepare a path. Only used by local_engine
        """

    @abc.abstractmethod
    def put_metadata(self, location, doc):
        """
        :param location Location: Location obj
        :param doc dict:
        :rtype: bool

        Upload a metadata doc. Metadata is different because of how indexes
        are stored.
        """

    @abc.abstractmethod
    def get_metadata(self, location):
        """
        :param location Location: Location obj
        :rtype: dict

        Download a metadata doc. Metadata is different because of how indexes
        are stored.
        """

    @abc.abstractmethod
    def close(self):
        """
        Delete all local cache to free memory
        """

    def find_one_id(self, prefix):
        """
        :param prefix Location: Location obj
        :param limit int:
        :rtype Location|None

        Convenience method for finding a single object given the prefix.
        Returns a Location if an object exists, otherwise, None
        """

        objs = self.list_ids(prefix, 1)
        if objs:
            return objs[0]



import datetime
import io
from typing import List, Tuple, Optional

import httpx
from bson.tz_util import utc
from fastapi import UploadFile
from starlette.datastructures import Headers

from retk import const, core
from retk.config import is_local_db
from retk.logger import logger
from retk.models.client import client
from retk.models.tps import AuthedUser
from retk.utils import ssrf_check, ASYNC_CLIENT_HEADERS
from .importing import async_tasks, sync_tasks

QUEUE_INITED = False


async def upload_obsidian(au: AuthedUser, zipped_files: List[UploadFile]) -> const.CodeEnum:
    max_file_count = 1
    max_file_size = 1024 * 1024 * 200  # 200 mb

    if len(zipped_files) > max_file_count:
        return const.CodeEnum.TOO_MANY_FILES

    zipped_file = zipped_files[0]
    filename = zipped_file.filename
    if not filename.endswith(".zip"):
        return const.CodeEnum.INVALID_FILE_TYPE

    if zipped_file.content_type not in ["application/zip", "application/octet-stream", "application/x-zip-compressed"]:
        return const.CodeEnum.INVALID_FILE_TYPE

    bytes_data = zipped_file.file.read()

    if is_local_db():
        # local db not support find_one_and_update
        await async_tasks.upload_obsidian_task(
            bytes_data=bytes_data,
            filename=filename,
            max_file_size=max_file_size,
            uid=au.u.id,
            request_id=au.request_id,
        )
    else:
        global QUEUE_INITED
        if not QUEUE_INITED:
            async_tasks.init()
            QUEUE_INITED = True
        async_tasks.QUEUE.put_nowait({
            "task": "obsidian",
            "bytes_data": bytes_data,
            "filename": filename,
            "max_file_size": max_file_size,
            "uid": au.u.id,
            "request_id": au.request_id,
        })
    return const.CodeEnum.OK


async def upload_text(au: AuthedUser, files: List[UploadFile]) -> const.CodeEnum:
    max_file_count = 200
    max_file_size = 1024 * 512  # 512 kb

    doc = await client.coll.import_data.find_one({"uid": au.u.id})
    if doc is not None and doc["running"]:
        return const.CodeEnum.IMPORT_PROCESS_NOT_FINISHED

    if len(files) > max_file_count:
        return const.CodeEnum.TOO_MANY_FILES

    file_list = [{
        "filename": file.filename,
        "content": file.file.read(),
        "size": file.size,
    } for file in files]

    if is_local_db():
        # local db not support find_one_and_update
        await async_tasks.update_text_task(
            files=file_list,
            max_file_size=max_file_size,
            uid=au.u.id,
            request_id=au.request_id,
        )
    else:
        global QUEUE_INITED
        if not QUEUE_INITED:
            async_tasks.init()
            QUEUE_INITED = True
        async_tasks.QUEUE.put_nowait({
            "task": "md",
            "files": file_list,
            "max_file_size": max_file_size,
            "uid": au.u.id,
            "request_id": au.request_id,
        })
    return const.CodeEnum.OK


async def get_upload_process(uid: str) -> Optional[dict]:
    timeout_minus = 5
    doc = await client.coll.import_data.find_one({"uid": uid})
    if doc is None:
        return None
    now = datetime.datetime.now(tz=utc)

    # upload timeout
    if doc["running"] and \
            now.replace(tzinfo=None) - doc["startAt"].replace(tzinfo=None) \
            > datetime.timedelta(minutes=timeout_minus):
        doc["running"] = False
        await client.coll.import_data.update_one(
            {"uid": uid},
            {"$set": {
                "running": False,
                "code": const.CodeEnum.UPLOAD_TASK_TIMEOUT.value,
                "msg": f"Timeout, upload not finish in {timeout_minus} mins",
            }},
        )
    return doc


async def vditor_upload(au: AuthedUser, files: List[UploadFile]) -> dict:
    res = {
        "errFiles": [],
        "succMap": {},
        "code": const.CodeEnum.OK,
    }
    if await core.user.user_space_not_enough(au=au):
        res["errFiles"] = [file.filename for file in files]
        res["code"] = const.CodeEnum.USER_SPACE_NOT_ENOUGH
        return res

    return await sync_tasks.save_editor_upload_files(
        uid=au.u.id,
        files=files,
    )


async def fetch_image_vditor(au: AuthedUser, url: str, count=0) -> Tuple[str, const.CodeEnum]:
    if count > 2:
        logger.debug(f"too many 30X code, failed to get {url}")
        return "", const.CodeEnum.FILE_OPEN_ERROR
    if ssrf_check(url):
        logger.debug(f"ssrf check failed: {url}")
        return "", const.CodeEnum.FILE_OPEN_ERROR
    if await core.user.user_space_not_enough(au=au):
        return "", const.CodeEnum.USER_SPACE_NOT_ENOUGH
    async with httpx.AsyncClient() as ac:
        try:
            response = await ac.get(
                url=url,
                headers=ASYNC_CLIENT_HEADERS,
                follow_redirects=False,
                timeout=5.
            )
        except (
                httpx.ConnectTimeout,
                RuntimeError,
                httpx.ConnectError,
                httpx.ReadTimeout,
                httpx.HTTPError
        ) as e:
            logger.debug(f"failed to get {url}: {e}")
            return "", const.CodeEnum.FILE_OPEN_ERROR
        if response.status_code in [301, 302]:
            return await fetch_image_vditor(au=au, url=response.headers["Location"], count=count + 1)
        elif response.status_code != 200:
            return "", const.CodeEnum.FILE_OPEN_ERROR

        content = response.content

        file = UploadFile(
            filename=url.split("?", 1)[0].split("/")[-1],  # remove url parameters
            file=io.BytesIO(content),
            headers=Headers(response.headers),
            size=len(content)
        )

    if not file.content_type.startswith(const.app.ValidUploadedFilePrefixEnum.IMAGE.value):
        return "", const.CodeEnum.INVALID_FILE_TYPE

    res = await sync_tasks.save_editor_upload_files(
        uid=au.u.id,
        files=[file],
    )
    if len(res["errFiles"]) > 0:
        return "", const.CodeEnum.FILE_OPEN_ERROR
    return res["succMap"][file.filename], const.CodeEnum.OK
